import warp as wp


@wp.func
def Subtraction(a: float, b: float) -> float:
  return wp.max(a, -b)


@wp.func
def Intersection(a: float, b: float) -> float:
  return wp.max(a, b)


@wp.func
def circle(rho: float, r: float) -> float:
  return rho - r


@wp.func
def smoothUnion(a: float, b: float, k: float) -> float:
  h = wp.min(wp.max(0.5 + 0.5 * (b - a) / k, 0.0), 1.0)
  return b * (1.0 - h) + a * h - k * h * (1.0 - h)


@wp.func
def smoothIntersection(a: float, b: float, k: float) -> float:
  return Subtraction(Intersection(a, b), smoothUnion(Subtraction(a, b), Subtraction(b, a), k))


@wp.func
def extrusion(p: wp.vec3, sdf_2d: float, h: float) -> float:
  w = wp.vec2()
  w[0] = sdf_2d
  w[1] = wp.abs(p[2]) - h
  w_abs = wp.vec2()
  w_abs[0] = wp.max(w[0], 0.0)
  w_abs[1] = wp.max(w[1], 0.0)
  return wp.min(wp.max(w[0], w[1]), 0.0) + wp.sqrt(w_abs[0] * w_abs[0] + w_abs[1] * w_abs[1])


@wp.func
def mod(x: float, y: float) -> float:
  return x - y * wp.floor(x / y)


@wp.func
def distance2D(p: wp.vec3, attributes: wp.vec3) -> float:
  # see https://www.shadertoy.com/view/3lG3WR
  D = 2.8
  N = 25.0
  psi = 3.096e-5 * N * N - 6.557e-3 * N + 0.551  # pressure angle
  alpha = 0.0
  innerdiameter = -1.0

  R = D / 2.0
  rho = wp.sqrt(p[0] * p[0] + p[1] * p[1])
  Pd = N / D  # Diametral Pitch: teeth per unit length of diameter
  P = wp.PI / Pd  # Circular Pitch
  a = 1.0 / Pd  # Addendum: radial length of a tooth from the pitch
  # circle to the tip of the tooth.

  Do = D + 2.0 * a  # Outside Diameter
  Ro = Do / 2.0

  h = 2.2 / Pd

  innerR = Ro - h - 0.14 * D
  if innerdiameter >= 0.0:
    innerR = innerdiameter / 2.0

  # Early exit
  if innerR - rho > 0.0:
    return innerR - rho

  # Early exit
  if Ro - rho < -0.2:
    return rho - Ro

  Db = D * wp.cos(psi)  # Base Diameter
  Rb = Db / 2.0

  fi = wp.atan2(p[1], p[0]) + alpha
  alphaStride = P / R

  invAlpha = wp.acos(Rb / R)
  invPhi = wp.tan(invAlpha) - invAlpha

  shift = alphaStride / 2.0 - 2.0 * invPhi

  fia = mod(fi + shift / 2.0, alphaStride) - shift / 2.0
  fib = mod(-fi - shift + shift / 2.0, alphaStride) - shift / 2.0

  dista = -1.0e6
  distb = -1.0e6

  if Rb < rho:
    acos_rbRho = wp.acos(Rb / rho)

    thetaa = fia + acos_rbRho
    thetab = fib + acos_rbRho

    ta = wp.sqrt(rho * rho - Rb * Rb)

    # https://math.stackexchange.com/questions/1266689/distance-from-a-point-to-the-involute-of-a-circle
    dista = ta - Rb * thetaa
    distb = ta - Rb * thetab

  gearOuter = circle(rho, Ro)
  gearLowBase = circle(rho, Ro - h)
  crownBase = circle(rho, innerR)
  cogs = Intersection(dista, distb)
  baseWalls = Intersection(fia - (alphaStride - shift), fib - (alphaStride - shift))

  cogs = Intersection(baseWalls, cogs)
  cogs = smoothIntersection(gearOuter, cogs, 0.0035 * D)
  cogs = smoothUnion(gearLowBase, cogs, Rb - Ro + h)
  cogs = Subtraction(cogs, crownBase)

  return cogs


@wp.func
def gear(p: wp.vec3, attr: wp.vec3) -> float:
  thickness = 0.2
  return extrusion(p, distance2D(p, attr), thickness / 2.0)


@wp.func
def gear_sdf_grad(p: wp.vec3, attr: wp.vec3) -> wp.vec3:
  grad = wp.vec3()
  eps = 1e-6
  f_original = gear(p, attr)
  x_plus = wp.vec3(p[0] + eps, p[1], p[2])
  f_plus = gear(x_plus, attr)
  grad[0] = (f_plus - f_original) / eps

  x_plus = wp.vec3(p[0], p[1] + eps, p[2])
  f_plus = gear(x_plus, attr)
  grad[1] = (f_plus - f_original) / eps

  x_plus = wp.vec3(p[0], p[1], p[2] + eps)
  f_plus = gear(x_plus, attr)
  grad[2] = (f_plus - f_original) / eps
  return grad



# Copyright 2025 The Newton Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Tests for sensor functions."""

import mujoco
import numpy as np
import warp as wp
from absl.testing import absltest
from absl.testing import parameterized

import mujoco_warp as mjw
from mujoco_warp import DisableBit
from mujoco_warp import GeomType
from mujoco_warp import test_data
from mujoco_warp._src.collision_driver import MJ_COLLISION_TABLE
from mujoco_warp._src.collision_driver import create_collision_context
from mujoco_warp._src.types import CollisionType

# tolerance for difference between MuJoCo and MJWarp calculations - mostly
# due to float precision
_TOLERANCE = 5e-5


def _assert_eq(a, b, name):
  tol = _TOLERANCE * 10  # avoid test noise
  err_msg = f"mismatch: {name}"
  np.testing.assert_allclose(a, b, err_msg=err_msg, atol=tol, rtol=tol)


class SensorTest(parameterized.TestCase):
  def setUp(self):
    super().setUp()

    ## initialize primitive colliders
    # clear cache
    mjw._src.collision_primitive._PRIMITIVE_COLLISION_TYPES.clear()
    mjw._src.collision_primitive._PRIMITIVE_COLLISION_FUNC.clear()

    # map enum to string
    geom_str = {
      GeomType.PLANE: "plane",
      GeomType.SPHERE: "sphere",
      GeomType.CAPSULE: "capsule",
      GeomType.ELLIPSOID: "ellipsoid",
      GeomType.CYLINDER: "cylinder",
      GeomType.BOX: "box",
      GeomType.MESH: "mesh",
    }

    # generate kernel
    for geomtype in mjw._src.collision_primitive._PRIMITIVE_COLLISIONS.keys():
      if geomtype[0] == GeomType.PLANE:
        obj0 = '<geom type="plane" size="10 10 .01"/>'
      else:
        obj0 = f"""
        <body>
          <geom type="{geom_str[geomtype[0]]}" size=".1 .1 .1"/>
          <joint/>
        </body>
      """

      if geomtype[1] == GeomType.MESH:
        obj1 = f"""
          <body>
            <geom type="{geom_str[geomtype[1]]}" mesh="mesh" size=".1 .1 .1"/>
            <joint/>
          </body>
        """
        mesh = """
        <asset>
          <mesh name="mesh" builtin="sphere" params="0"/>
        </asset>
        """
      else:
        obj1 = f"""
          <body>
            <geom type="{geom_str[geomtype[1]]}" size=".1 .1 .1"/>
            <joint/>
          </body>
        """
        mesh = ""

      _, _, m, d = test_data.fixture(
        xml=f"""
      <mujoco>
        {mesh}
        <worldbody>
        {obj0}
        {obj1}
        </worldbody>
      </mujoco>
      """
      )

      ctx = create_collision_context(d.naconmax)
      primitive_pairs = [key for key, value in MJ_COLLISION_TABLE.items() if value == CollisionType.PRIMITIVE]
      mjw.primitive_narrowphase(m, d, ctx, primitive_pairs)

  def test_sensor(self):
    """Test sensors."""
    _, mjd, m, d = test_data.fixture(
      xml="""
      <mujoco>
        <option gravity="-1 -1 -1"/>
        <worldbody>
          <body name="body0" pos="0.1 0.2 0.3" quat=".05 .1 .15 .2">
            <joint name="slide" type="slide"/>
            <geom name="geom0" type="sphere" size="0.1"/>
            <site name="site0"/>
            <camera name="cam0"/>
          </body>
          <body name="body1" pos=".5 .6 .7">
            <joint name="ballquat" type="ball"/>
            <geom type="sphere" size="0.1"/>
            <site name="site1" pos=".1 .2 .3"/>
          </body>
          <body name="body2" pos="1 1 1">
            <freejoint/>
            <geom type="sphere" size="0.1"/>
            <site name="site2"/>
          </body>
          <body name="body3" pos="2 2 2">
            <joint name="hinge0" type="hinge" axis="1 0 0"/>
            <geom type="sphere" size="0.1" pos=".1 0 0"/>
            <body pos="2 2 2">
              <joint name="hinge1" type="hinge" axis="1 0 0"/>
              <geom type="sphere" size="0.1" pos=".1 0 0"/>
            </body>
          </body>
          <body name="body4" pos="1 0 0">
            <joint type="ball"/>
            <geom type="sphere" size="0.1" pos=".1 0 0"/>
            <body>
              <joint type="ball"/>
              <geom type="sphere" size="0.1" pos=".1 0 0"/>
            </body>
          </body>
          <body pos="10 0 0">
            <joint type="hinge" axis="1 2 3"/>
            <geom type="sphere" size="0.1"/>
            <site name="force_site" pos="1 2 3"/>
          </body>
          <body pos="20 0 0">
            <joint type="slide" axis="1 2 3"/>
            <geom type="sphere" size="0.1"/>
            <site name="torque_site" pos="1 2 3"/>
          </body>
          <body name="body8">
            <joint type="hinge"/>
            <geom type="sphere" size="0.1" pos="1 2 3"/>
            <body name="body9">
              <joint type="hinge"/>
              <geom name="geom9" type="sphere" size="0.1" pos="1 2 3"/>
              <site name="site9" pos=".2 .4 .6"/>
            </body>
          </body>
          <camera name="camera"/>
          <site name="camera_site" pos="0 0 -1"/>
          <!-- limit pos: slide -->
          <body pos="1 1 1">
            <joint name="limitslide" type="slide" limited="true" range="-.5 .5" margin=".1"/>
            <geom type="sphere" size=".1"/>
          </body>
          <!-- limit pos: hinge -->
          <body pos="2 2 2">
            <joint name="limithinge" type="hinge" limited="true" range="-.4 .4" margin=".09"/>
            <geom type="sphere" size=".1"/>
          </body>
          <!-- limit pos: ball -->
          <body pos="3 3 3">
            <joint name="limitball" type="ball" limited="true" range="0 .1" margin=".05"/>
            <geom type="sphere" size=".1"/>
          </body>
          <!-- tendon limit pos -->
          <site name="sitetendon0" pos="5 4 4"/>
          <body pos="4 4 4">
            <joint type="slide" axis="1 0 0"/>
            <geom type="sphere" size=".1"/>
            <site name="sitetendon1"/>
          </body>
          <body pos="3 3 3">
            <joint name="tendonjoint" type="hinge"/>
            <geom type="sphere" size=".1"/>
          </body>
        </worldbody>
        <tendon>
          <spatial name="limittendon" limited="true" range="0 .5" margin=".1">
            <site site="sitetendon0"/>
            <site site="sitetendon1"/>
          </spatial>
          <fixed name="tendon">
            <joint joint="tendonjoint" coef=".123"/>
          </fixed>
        </tendon>
        <actuator>
          <motor name="slide" joint="slide"/>
          <motor tendon="tendon"/>
          <motor tendon="tendon" gear="2"/>
        </actuator>
        <sensor>
          <magnetometer site="site0"/>
          <camprojection camera="camera" site="camera_site"/>
          <camprojection camera="camera" site="camera_site" cutoff=".001"/>
          <jointpos joint="slide"/>
          <jointpos joint="slide" cutoff=".001"/>
          <actuatorpos actuator="slide"/>
          <actuatorpos actuator="slide" cutoff=".001"/>
          <ballquat joint="ballquat"/>
          <jointlimitpos joint="limitslide"/>
          <jointlimitpos joint="limitslide" cutoff=".001"/>
          <jointlimitpos joint="limithinge"/>
          <jointlimitpos joint="limithinge" cutoff=".001"/>
          <jointlimitpos joint="limitball"/>
          <jointlimitpos joint="limitball" cutoff=".001"/>
          <tendonlimitpos tendon="limittendon"/>
          <tendonlimitpos tendon="limittendon" cutoff=".001"/>
          <framepos objtype="body" objname="body1"/>
          <framepos objtype="body" objname="body1" cutoff=".001"/>
          <framepos objtype="body" objname="body1" reftype="body" refname="body0"/>
          <framepos objtype="body" objname="body1" reftype="geom" refname="geom0"/>
          <framepos objtype="body" objname="body1" reftype="site" refname="site0"/>
          <framepos objtype="body" objname="body1" reftype="cam" refname="cam0"/>
          <framepos objtype="xbody" objname="body1"/>
          <framepos objtype="geom" objname="geom0"/>
          <framepos objtype="site" objname="site0"/>
          <framepos objtype="camera" objname="cam0"/>
          <framexaxis objtype="body" objname="body1"/>
          <framexaxis objtype="body" objname="body1" reftype="body" refname="body0"/>
          <framexaxis objtype="body" objname="body1" reftype="geom" refname="geom0"/>
          <framexaxis objtype="body" objname="body1" reftype="site" refname="site0"/>
          <framexaxis objtype="body" objname="body1" reftype="cam" refname="cam0"/>
          <framexaxis objtype="xbody" objname="body1"/>
          <framexaxis objtype="geom" objname="geom0"/>
          <framexaxis objtype="site" objname="site0"/>
          <framexaxis objtype="camera" objname="cam0"/>
          <frameyaxis objtype="body" objname="body1"/>
          <frameyaxis objtype="body" objname="body1" reftype="body" refname="body0"/>
          <frameyaxis objtype="body" objname="body1" reftype="geom" refname="geom0"/>
          <frameyaxis objtype="body" objname="body1" reftype="site" refname="site0"/>
          <frameyaxis objtype="body" objname="body1" reftype="cam" refname="cam0"/>
          <frameyaxis objtype="xbody" objname="body1"/>
          <frameyaxis objtype="geom" objname="geom0"/>
          <frameyaxis objtype="site" objname="site0"/>
          <frameyaxis objtype="camera" objname="cam0"/>
          <framezaxis objtype="body" objname="body1"/>
          <framezaxis objtype="body" objname="body1" reftype="body" refname="body0"/>
          <framezaxis objtype="body" objname="body1" reftype="geom" refname="geom0"/>
          <framezaxis objtype="body" objname="body1" reftype="site" refname="site0"/>
          <framezaxis objtype="body" objname="body1" reftype="cam" refname="cam0"/>
          <framezaxis objtype="xbody" objname="body1"/>
          <framezaxis objtype="geom" objname="geom0"/>
          <framezaxis objtype="site" objname="site0"/>
          <framezaxis objtype="camera" objname="cam0"/>
          <framequat objtype="body" objname="body1"/>
          <framequat objtype="body" objname="body1" reftype="body" refname="body0"/>
          <framequat objtype="body" objname="body1" reftype="geom" refname="geom0"/>
          <framequat objtype="body" objname="body1" reftype="site" refname="site0"/>
          <framequat objtype="body" objname="body1" reftype="cam" refname="cam0"/>
          <framequat objtype="xbody" objname="body1"/>
          <framequat objtype="geom" objname="geom0"/>
          <framequat objtype="site" objname="site0"/>
          <framequat objtype="camera" objname="cam0"/>
          <subtreecom body="body3"/>
          <subtreecom body="body3" cutoff=".001"/>
          <e_potential/>
          <e_potential cutoff=".001"/>
          <e_kinetic/>
          <e_kinetic cutoff=".001"/>
          <clock/>
          <clock cutoff=".001"/>
          <velocimeter site="site2"/>
          <velocimeter site="site2" cutoff=".001"/>
          <gyro site="site2"/>
          <gyro site="site2" cutoff=".001"/>
          <jointvel joint="slide"/>
          <jointvel joint="slide" cutoff=".001"/>
          <actuatorvel actuator="slide"/>
          <actuatorvel actuator="slide" cutoff=".001"/>
          <ballangvel joint="ballquat"/>
          <ballangvel joint="ballquat" cutoff=".001"/>
          <jointlimitvel joint="limithinge"/>
          <jointlimitvel joint="limithinge" cutoff=".001"/>
          <tendonlimitvel tendon="limittendon"/>
          <tendonlimitvel tendon="limittendon" cutoff=".001"/>
          <framelinvel objtype="body" objname="body9"/>
          <framelinvel objtype="body" objname="body9" cutoff=".001"/>
          <frameangvel objtype="body" objname="body9"/>
          <frameangvel objtype="body" objname="body9" cutoff=".001"/>
          <framelinvel objtype="xbody" objname="body9"/>
          <frameangvel objtype="xbody" objname="body9"/>
          <framelinvel objtype="geom" objname="geom9"/>
          <frameangvel objtype="geom" objname="geom9"/>
          <framelinvel objtype="site" objname="site9"/>
          <frameangvel objtype="site" objname="site9"/>
          <framelinvel objtype="camera" objname="cam0"/>
          <frameangvel objtype="camera" objname="cam0"/>
          <framelinvel objtype="body" objname="body9" reftype="xbody" refname="body0"/>
          <frameangvel objtype="body" objname="body9" reftype="xbody" refname="body0"/>
          <framelinvel objtype="body" objname="body9" reftype="geom" refname="geom0"/>
          <frameangvel objtype="body" objname="body9" reftype="geom" refname="geom0"/>
          <framelinvel objtype="body" objname="body9" reftype="site" refname="site0"/>
          <frameangvel objtype="body" objname="body9" reftype="site" refname="site0"/>
          <framelinvel objtype="body" objname="body9" reftype="cam" refname="cam0"/>
          <frameangvel objtype="body" objname="body9" reftype="cam" refname="cam0"/>
          <subtreelinvel body="body4"/>
          <subtreelinvel body="body4" cutoff=".001"/>
          <subtreeangmom body="body4"/>
          <subtreeangmom body="body4" cutoff=".001"/>
          <accelerometer site="force_site"/>
          <accelerometer site="force_site" cutoff=".001"/>
          <force site="force_site"/>
          <force site="force_site" cutoff=".001"/>
          <torque site="torque_site"/>
          <torque site="torque_site" cutoff=".001"/>
          <actuatorfrc actuator="slide"/>
          <actuatorfrc actuator="slide" cutoff=".001"/>
          <jointactuatorfrc joint="slide"/>
          <jointactuatorfrc joint="slide" cutoff=".001"/>
          <jointlimitfrc joint="limitslide"/>
          <jointlimitfrc joint="limitslide" cutoff=".001"/>
          <jointlimitfrc joint="limithinge"/>
          <jointlimitfrc joint="limithinge" cutoff=".001"/>
          <jointlimitfrc joint="limitball"/>
          <jointlimitfrc joint="limitball" cutoff=".001"/>
          <tendonlimitfrc tendon="limittendon"/>
          <tendonlimitfrc tendon="limittendon" cutoff=".001"/>
          <tendonactuatorfrc tendon="tendon"/>
          <tendonactuatorfrc tendon="tendon" cutoff=".001"/>
          <framelinacc objtype="body" objname="body9"/>
          <framelinacc objtype="body" objname="body9" cutoff=".001"/>
          <frameangacc objtype="body" objname="body9"/>
          <frameangacc objtype="body" objname="body9" cutoff=".001"/>
          <framelinacc objtype="xbody" objname="body9"/>
          <frameangacc objtype="xbody" objname="body9"/>
          <framelinacc objtype="geom" objname="geom9"/>
          <frameangacc objtype="geom" objname="geom9"/>
          <framelinacc objtype="site" objname="site9"/>
          <frameangacc objtype="site" objname="site9"/>
          <framelinacc objtype="camera" objname="cam0"/>
          <frameangacc objtype="camera" objname="cam0"/>
        </sensor>
        <keyframe>
          <key qpos="1 .1 .2 .3 .4 1 1 1 1 0 0 0 .25 .35 1 0 0 0 1 0 0 0 0 0 1 1 .6 .5 1 2 3 4 .5 0" qvel="2 .2 -.1 .4 .25 .35 .45 -0.1 -0.2 -0.3 .1 -.2 -.5 -0.75 -1 .1 .2 .3 0 0 2 2 0 0 0 0 0 0 0" ctrl="3 .1 .2"/>
        </keyframe>
      </mujoco>
    """,
      keyframe=0,
      qvel_noise=0.01,
      ctrl_noise=0.1,
    )

    d.sensordata.zero_()

    mjw.sensor_pos(m, d)
    mjw.sensor_vel(m, d)
    mjw.sensor_acc(m, d)

    _assert_eq(d.sensordata.numpy()[0], mjd.sensordata, "sensordata")

  def test_rangefinder(self):
    """Test rangefinder."""
    for keyframe in range(2):
      _, mjd, m, d = test_data.fixture(
        xml="""
        <mujoco>
          <compiler angle="degree"/>
          <worldbody>
            <geom type="sphere" size=".1" pos="0 0 1"/>
            <body>
              <joint name="joint0" type="hinge" axis="1 0 0"/>
              <joint type="hinge" axis="0 1 0"/>
              <joint type="hinge" axis="0 0 1"/>
              <geom type="sphere" size="0.1"/>
              <site name="site0" size=".1"/>
              <site name="site1" size=".1" pos="0 0 -.2"/>
            </body>
          </worldbody>
          <sensor>
            <rangefinder site="site0"/>
            <jointpos joint="joint0"/>
            <rangefinder site="site1"/>
          </sensor>
          <keyframe>
            <key qpos="0 0 0"/>
            <key qpos="0 90 0"/>
          </keyframe>
        </mujoco>
      """,
        keyframe=keyframe,
      )

      d.sensordata.zero_()

      mjw.sensor_pos(m, d)

      _assert_eq(d.sensordata.numpy()[0], mjd.sensordata, "sensordata")

  def test_touch_sensor(self):
    """Test touch sensor."""
    for keyframe in range(2):
      _, mjd, m, d = test_data.fixture(
        xml="""
        <mujoco>
          <worldbody>
            <geom type="plane" size="10 10 .001"/>
            <body pos="0 0 .25">
              <geom type="sphere" size="0.1" pos=".1 0 0"/>
              <geom type="sphere" size="0.1" pos="-.1 0 0"/>
              <geom type="sphere" size="0.1" pos="0 0 .11"/>
              <geom type="sphere" size="0.1" pos="0 0 10"/>
              <site name="site_sphere" type="sphere" size=".2"/>
              <site name="site_capsule" type="capsule" size=".2 .2"/>
              <site name="site_ellipsoid" type="ellipsoid" size=".2 .2 .2"/>
              <site name="site_cylinder" type="cylinder" size=".2 .2"/>
              <site name="site_box" type="box" size=".2 .2 .2"/>
              <freejoint/>
            </body>
          </worldbody>
          <sensor>
            <touch site="site_sphere"/>
            <touch site="site_capsule"/>
            <touch site="site_ellipsoid"/>
            <touch site="site_cylinder"/>
            <touch site="site_box"/>
          </sensor>
          <keyframe>
            <key qpos="0 0 10 1 0 0 0"/>
            <key qpos="0 0 .05 1 0 0 0"/>
            <key qpos="0 0 0 1 0 0 0"/>
          </keyframe>
        </mujoco>
      """,
        keyframe=keyframe,
      )

      d.sensordata.zero_()

      mjw.sensor_acc(m, d)

      _assert_eq(d.sensordata.numpy()[0], mjd.sensordata, "sensordata")

  def test_tendon_sensor(self):
    """Test tendon sensors."""
    _, mjd, m, d = test_data.fixture("tendon/fixed.xml", keyframe=0)

    d.sensordata.zero_()

    mjw.sensor_pos(m, d)
    mjw.sensor_vel(m, d)

    _assert_eq(d.sensordata.numpy()[0], mjd.sensordata, "sensordata")

  @parameterized.parameters("humanoid/humanoid.xml", "constraints.xml")
  def test_energy(self, xml):
    mjm, mjd, m, d = test_data.fixture(
      xml, qvel_noise=0.01, ctrl_noise=0.1, overrides={"opt.disableflags": DisableBit.CONSTRAINT}
    )

    d.energy.zero_()

    mujoco.mj_energyPos(mjm, mjd)
    mjw.energy_pos(m, d)
    _assert_eq(d.energy.numpy()[0][0], mjd.energy[0], "potential energy")

    mujoco.mj_energyVel(mjm, mjd)
    mjw.energy_vel(m, d)
    _assert_eq(d.energy.numpy()[0][1], mjd.energy[1], "kinetic energy")

  @parameterized.parameters(
    'type="capsule" size=".1 .1" euler="0 89 89"',
    'type="box" size=".1 .11 .12" euler=".02 .05 .1"',
  )
  def test_contact_sensor(self, geom):
    """Test contact sensor."""
    # create contact sensors
    contact_sensor = ""

    # data combinations
    datas = ("found", "force dist normal", "torque pos tangent", "found force torque dist pos normal tangent")

    for geoms in (
      'geom2="plane"',
      'geom1="plane" geom2="geom"',
      'geom1="geom" geom2="plane"',
      'body1="plane"',
      'body1="plane" body2="geom"',
      'body1="geom" body2="plane"',
    ):
      for num in (1, 5):
        for reduce in (None, "mindist", "maxforce"):
          for data in datas:
            contact_sensor += f'<contact {geoms} num="{num}"'
            if reduce is not None:
              contact_sensor += f' reduce="{reduce}"'
            contact_sensor += f' data="{data}"/>'

    _MJCF = f"""
    <mujoco>
      <compiler angle="degree"/>
      <option cone="pyramidal"/>
      <worldbody>
        <body name="plane">
          <geom name="plane" type="plane" size="10 10 .001"/>
        </body>
        <body name="geom">
          <geom name="geom" {geom}/>
          <joint type="slide" axis="0 0 1"/>
        </body>
        <body name="sphere">
          <geom name="sphere" type="sphere" size=".1"/>
          <joint type="slide" axis="0 0 1"/>
        </body>
      </worldbody>
      <keyframe>
        <key qpos=".08 1"/>
      </keyframe>
      <sensor>
        {contact_sensor}
      </sensor>
    </mujoco>
    """

    _, mjd, m, d = test_data.fixture(xml=_MJCF, keyframe=0)

    d.sensordata.zero_()
    mjw.sensor_acc(m, d)

    sensordata = d.sensordata.numpy()[0]
    _assert_eq(sensordata, mjd.sensordata, "sensordata")
    self.assertTrue(sensordata.any())  # check that sensordata is not empty

  def test_contact_sensor_subtree(self):
    """Test contact sensor with subtree matching semantics."""
    _MJCF = """
    <mujoco>
      <worldbody>
        <geom type="plane" size="2 2 .01"/>
        <body name="thigh" pos="-1 0 .1">
          <joint type="slide"/>
          <geom type="capsule" size=".1" fromto="0 0 0 .5 0 0"/>
          <body name="shin" pos=".7 0 0">
            <joint axis="0 1 0"/>
            <geom type="capsule" size=".1" fromto="0 0 0 0 0 .5"/>
            <body name="foot" pos="0 0 .7">
              <joint axis="0 1 0"/>
              <geom type="capsule" size=".1" fromto="0 0 0 -.7 0 0"/>
            </body>
          </body>
        </body>
      </worldbody>
      <sensor>
        <contact name="all" reduce="mindist"/>
        <contact name="world" subtree1="world" reduce="mindist"/>
        <contact name="thigh" subtree1="thigh" reduce="mindist"/>
        <contact name="shin" subtree1="shin" reduce="mindist"/>
        <contact name="foot" subtree1="foot" reduce="mindist"/>
        <contact name="foot_w" subtree1="foot" body2="world" reduce="mindist"/>
        <contact name="foot_w2" subtree1="foot" subtree2="world" reduce="mindist"/>
      </sensor>
      <keyframe>
        <key qpos="-6.96651e-05 -0.0478055 -0.746498"/>
      </keyframe>
    </mujoco>
    """
    _, _, m, d = test_data.fixture(xml=_MJCF, keyframe=0)

    d.sensordata.zero_()
    mjw.sensor_acc(m, d)

    _assert_eq(d.sensordata.numpy()[0], np.array([4, 4, 4, 2, 1, 0, 1]), "found")

  def test_contact_sensor_subtree_found_zero(self):
    """Test contact sensor with subtree matching semantics generates found=0."""
    _MJCF = """
    <mujoco>
      <worldbody>
        <geom type="plane" size="2 2 .01"/>
        <body name="base">
          <joint type="slide" axis="1 0 0"/>
          <joint type="slide" axis="0 0 1"/>
          <geom type="capsule" size=".1" fromto="0 0 0 1 0 0"/>
          <body name="body0">
            <joint axis="0 1 0"/>
            <geom type="capsule" size=".1" fromto="0 0 0 0 0 -1"/>
          </body>
          <body name="body1" pos="1 0 0">
            <joint axis="0 1 0"/>
            <geom type="capsule" size=".1" fromto="0 0 0 0 0 -1"/>
          </body>
        </body>
      </worldbody>
      <sensor>
        <contact subtree1="base" subtree2="base"/>
      </sensor>
      <keyframe>
        <key qpos="0 1 0 0"/>
      </keyframe>
    </mujoco>
    """
    _, _, m, d = test_data.fixture(xml=_MJCF, keyframe=0)

    d.sensordata.fill_(wp.inf)
    mjw.sensor_acc(m, d)

    _assert_eq(d.nacon.numpy()[0], 2, "nacon")
    _assert_eq(d.sensordata.numpy()[0], 0, "found")

  @parameterized.product(site_geom=["sphere", "capsule", "ellipsoid", "cylinder", "box"], key_pos=["0 0 10", "0 0 .09"])
  def test_contact_sensor_site(self, site_geom, key_pos):
    _, mjd, m, d = test_data.fixture(
      xml=f"""
    <mujoco>
      <worldbody>
        <geom type="plane" size="10 10 .001"/>
        <site name="site" type="{site_geom}" size=".1 .2 .3"/>
        <body>
          <geom type="sphere" size=".1"/>
          <freejoint/>
        </body>
      </worldbody>
      <sensor>
        <contact site="site" reduce="mindist"/>
      </sensor>
      <keyframe>
        <key qpos="{key_pos} 1 0 0 0"/>
      </keyframe>
    </mujoco>
    """,
      keyframe=0,
    )

    d.sensordata.zero_()
    mjw.sensor_acc(m, d)

    _assert_eq(d.sensordata.numpy()[0], mjd.sensordata, "sensordata")

  def test_contact_sensor_netforce(self):
    """Test contact sensor with netforce reduction."""
    _, mjd, m, d = test_data.fixture(
      xml="""
    <mujoco>
      <worldbody>
        <geom name="plane" type="plane" size="10 10 .001"/>
        <body>
          <geom name="box" type="box" size=".1 .1 .1"/>
          <freejoint/>
        </body>
      </worldbody>
      <sensor>
        <contact geom1="plane" geom2="box" data="found force torque dist pos normal tangent" reduce="netforce" num="2"/>
        <contact geom1="box" geom2="plane" data="force torque" reduce="netforce"/>
      </sensor>
      <keyframe>
        <key qpos="0 0 .09 1 0 0 0"/>
      </keyframe>
    </mujoco>
    """,
      keyframe=0,
    )

    d.sensordata.zero_()
    mjw.sensor_acc(m, d)
    sensordata = d.sensordata.numpy()[0]
    _assert_eq(sensordata, mjd.sensordata, "sensordata")
    self.assertTrue(sensordata.any())  # check that sensordata is not empty

  @parameterized.parameters(
    ("box", "box", "box", "box"),
    ("sphere", "capsule", "ellipsoid", "cylinder"),
    ("capsule", "box", "cylinder", "sphere"),
    ("capsule", "cylinder", "box", "ellipsoid"),
    ("cylinder", "box", "ellipsoid", "capsule"),
  )
  def test_sensor_collision(self, type0, type1, type2, type3):
    """Tests collision sensors: distance, normal, fromto."""
    _MJCF = f"""
      <mujoco>
        <worldbody>
          <body name="obj0">
            <geom name="obj0" type="{type0}" size=".1 .1 .1" euler="1 2 3"/>
          </body>
          <body name="obj1" pos="0 0 1">
            <geom name="obj1" type="{type1}" size=".1 .1 .1" euler="-1 2 -1"/>
          </body>
          <body name="objobj" pos="0 0 -1">
            <geom name="objobj0" pos=".01 0 0.005" type="{type2}" size=".09 .09 .09" euler="2 1 3"/>
            <geom name="objobj1" pos="-.01 0 -0.0025" type="{type3}" size=".11 .11 .11" euler="3 1 2"/>
          </body>
        </worldbody>
        <sensor>
          <!-- distance geom-geom -->
          <distance geom1="obj0" geom2="obj1" cutoff="0"/>
          <distance geom1="obj0" geom2="obj1" cutoff="10"/>
          <distance geom1="obj1" geom2="obj0" cutoff="0"/>
          <distance geom1="obj1" geom2="obj0" cutoff="10"/>
          <distance body1="obj0" body2="obj1" cutoff="10"/>
          <distance body1="obj1" body2="obj0" cutoff="10"/>
          <distance body1="obj0" body2="obj1" cutoff="10"/>
          <distance body1="obj1" body2="obj0" cutoff="10"/>
          <distance geom1="obj0" body2="obj1" cutoff="10"/>
          <distance body1="obj1" geom2="obj0" cutoff="10"/>

          <!-- normal geom-geom -->
          <normal geom1="obj0" geom2="obj1" cutoff="0"/>
          <normal geom1="obj0" geom2="obj1" cutoff="10"/>
          <normal geom1="obj1" geom2="obj0" cutoff="0"/>
          <normal geom1="obj1" geom2="obj0" cutoff="10"/>
          <normal body1="obj0" body2="obj1" cutoff="10"/>
          <normal body1="obj1" body2="obj0" cutoff="10"/>
          <normal body1="obj0" body2="obj1" cutoff="10"/>
          <normal body1="obj1" body2="obj0" cutoff="10"/>
          <normal geom1="obj0" body2="obj1" cutoff="10"/>
          <normal body1="obj1" geom2="obj0" cutoff="10"/>

          <!-- fromto geom-geom -->
          <fromto geom1="obj0" geom2="obj1" cutoff="0"/>
          <fromto geom1="obj0" geom2="obj1" cutoff="10"/>
          <fromto geom1="obj1" geom2="obj0" cutoff="0"/>
          <fromto geom1="obj1" geom2="obj0" cutoff="10"/>
          <fromto body1="obj0" body2="obj1" cutoff="10"/>
          <fromto body1="obj1" body2="obj0" cutoff="10"/>
          <fromto body1="obj0" body2="obj1" cutoff="10"/>
          <fromto body1="obj1" body2="obj0" cutoff="10"/>
          <fromto geom1="obj0" body2="obj1" cutoff="10"/>
          <fromto body1="obj1" geom2="obj0" cutoff="10"/>

          <!-- distance geom body -->
          <distance geom1="obj0" body2="objobj" cutoff="0"/>
          <distance geom1="obj0" body2="objobj" cutoff="10"/>
          <distance body1="objobj" geom2="obj0" cutoff="0"/>
          <distance body1="objobj" geom2="obj0" cutoff="10"/>
          <distance body1="obj0" body2="objobj" cutoff="10"/>
          <distance body1="objobj" body2="obj0" cutoff="10"/>
          <distance body1="obj0" body2="objobj" cutoff="10"/>
          <distance body1="objobj" body2="obj0" cutoff="10"/>
          <distance geom1="obj0" body2="objobj" cutoff="10"/>
          <distance body1="objobj" geom2="obj0" cutoff="10"/>

          <!-- normal geom body -->
          <normal geom1="obj0" body2="objobj" cutoff="0"/>
          <normal geom1="obj0" body2="objobj" cutoff="10"/>
          <normal body1="objobj" geom2="obj0" cutoff="0"/>
          <normal body1="objobj" geom2="obj0" cutoff="10"/>
          <normal body1="obj0" body2="objobj" cutoff="10"/>
          <normal body1="objobj" body2="obj0" cutoff="10"/>
          <normal body1="obj0" body2="objobj" cutoff="10"/>
          <normal body1="objobj" body2="obj0" cutoff="10"/>
          <normal geom1="obj0" body2="objobj" cutoff="10"/>
          <normal body1="objobj" geom2="obj0" cutoff="10"/>

          <!-- fromto geom body -->
          <fromto geom1="obj0" body2="objobj" cutoff="0"/>
          <fromto geom1="obj0" body2="objobj" cutoff="10"/>
          <fromto body1="objobj" geom2="obj0" cutoff="0"/>
          <fromto body1="objobj" geom2="obj0" cutoff="10"/>
          <fromto body1="obj0" body2="objobj" cutoff="10"/>
          <fromto body1="objobj" body2="obj0" cutoff="10"/>
          <fromto body1="obj0" body2="objobj" cutoff="10"/>
          <fromto body1="objobj" body2="obj0" cutoff="10"/>
          <fromto geom1="obj0" body2="objobj" cutoff="10"/>
          <fromto body1="objobj" geom2="obj0" cutoff="10"/>
        </sensor>
      </mujoco>
      """

    _, mjd, m, d = test_data.fixture(xml=_MJCF)

    d.sensordata.fill_(wp.inf)
    mjw.collision(m, d)
    mjw.sensor_pos(m, d)

    _assert_eq(d.sensordata.numpy()[0], mjd.sensordata, "sensordata")

  @parameterized.parameters("sphere", "capsule", "ellipsoid", "cylinder", "box")
  def test_sensor_collision_plane(self, type_):
    """Tests collision sensors: distance, normal, fromto."""
    _MJCF = f"""
      <mujoco>
        <worldbody>
          <geom name="plane" type="plane" size="10 10 .01" euler="2 2 2"/>
          <body name="obj" pos="0 0 1">
            <geom name="obj" type="{type_}" size=".1 .1 .1" euler="1 2 3"/>
          </body>
        </worldbody>
        <sensor>
          <!-- distance geom-geom -->
          <distance geom1="plane" geom2="obj" cutoff="0"/>
          <distance geom1="plane" geom2="obj" cutoff="10"/>
          <distance geom1="obj" geom2="plane" cutoff="0"/>
          <distance geom1="obj" geom2="plane" cutoff="10"/>
          <distance geom1="plane" body2="obj" cutoff="10"/>

          <!-- normal geom-geom -->
          <normal geom1="plane" geom2="obj" cutoff="0"/>
          <normal geom1="plane" geom2="obj" cutoff="10"/>
          <normal geom1="obj" geom2="plane" cutoff="0"/>
          <normal geom1="obj" geom2="plane" cutoff="10"/>
          <normal geom1="plane" body2="obj" cutoff="10"/>

          <!-- fromto geom-geom -->
          <fromto geom1="plane" geom2="obj" cutoff="0"/>
          <fromto geom1="plane" geom2="obj" cutoff="10"/>
          <fromto geom1="obj" geom2="plane" cutoff="0"/>
          <fromto geom1="obj" geom2="plane" cutoff="10"/>
          <fromto geom1="plane" body2="obj" cutoff="10"/>
        </sensor>
      </mujoco>
      """

    _, mjd, m, d = test_data.fixture(xml=_MJCF)

    d.sensordata.fill_(wp.inf)
    mjw.collision(m, d)
    mjw.sensor_pos(m, d)

    _assert_eq(d.sensordata.numpy()[0], mjd.sensordata, "sensordata")

  @parameterized.parameters(0, 1)
  def test_insidesite(self, keyframe):
    _, mjd, m, d = test_data.fixture(
      xml="""
    <mujoco>
      <worldbody>
        <site name="refsite" type="sphere" size=".1"/>
        <body name="body">
          <geom name="geom" type="sphere" size=".1"/>
          <site name="site"/>
          <camera name="camera"/>
          <joint type="slide" axis="1 0 0"/>
        </body>
      </worldbody>
      <sensor>
        <insidesite site="refsite" objtype="xbody" objname="body"/>
        <insidesite site="refsite" objtype="body" objname="body"/>
        <insidesite site="refsite" objtype="geom" objname="geom"/>
        <insidesite site="refsite" objtype="site" objname="site"/>
        <insidesite site="refsite" objtype="camera" objname="camera"/>
      </sensor>
      <keyframe>
        <key qpos="0"/>
        <key qpos="1"/>
      </keyframe>
    </mujoco>
    """,
      keyframe=keyframe,
    )

    d.sensordata.fill_(wp.inf)
    mjw.sensor_pos(m, d)

    _assert_eq(d.sensordata.numpy()[0], mjd.sensordata, "sensordata")


if __name__ == "__main__":
  wp.init()
  absltest.main()



# Copyright 2025 The Newton Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for ray functions."""

import mujoco
import numpy as np
import warp as wp
from absl.testing import absltest

import mujoco_warp as mjw
from mujoco_warp import test_data
from mujoco_warp._src.types import vec6

# tolerance for difference between MuJoCo and MJX ray calculations - mostly
# due to float precision
_TOLERANCE = 5e-5


def _assert_eq(a, b, name):
  tol = _TOLERANCE * 10  # avoid test noise
  err_msg = f"mismatch: {name}"
  np.testing.assert_allclose(a, b, err_msg=err_msg, atol=tol, rtol=tol)


# TODO: Add tests comparing normal to engine implementation once available.
class RayTest(absltest.TestCase):
  def test_ray_nothing(self):
    """Tests that ray returns -1 when nothing is hit."""
    mjm, mjd, m, d = test_data.fixture("ray.xml")

    pnt = wp.array([wp.vec3(12.146, 1.865, 3.895)], dtype=wp.vec3).reshape((1, 1))
    vec = wp.array([wp.vec3(0.0, 0.0, -1.0)], dtype=wp.vec3).reshape((1, 1))
    dist, geomid, normal = mjw.ray(m, d, pnt, vec)
    wp.synchronize()
    geomid_np = geomid.numpy()[0, 0]  # Extract from [[-1]]
    dist_np = dist.numpy()[0, 0]  # Extract from [[-1.]]
    normal_np = normal.numpy()[0, 0]
    _assert_eq(geomid_np, -1, "geom_id")
    _assert_eq(dist_np, -1, "dist")
    _assert_eq(normal_np, 0, "normal")

    # test that bvh accelerated produces the same results
    rc = mjw.create_render_context(mjm)
    dist, geomid, normal = mjw.ray(m, d, pnt, vec, rc=rc)
    wp.synchronize()
    bvh_geomid_np = geomid.numpy()[0, 0]
    bvh_dist_np = dist.numpy()[0, 0]
    bvh_normal_np = normal.numpy()[0, 0]
    _assert_eq(bvh_geomid_np, geomid_np, "geom_id")
    _assert_eq(bvh_dist_np, dist_np, "dist")
    _assert_eq(bvh_normal_np, normal_np, "normal")

  def test_ray_plane(self):
    """Tests ray<>plane matches MuJoCo."""
    mjm, mjd, m, d = test_data.fixture("ray.xml")

    # looking down at a slight angle
    pnt = wp.array([wp.vec3(2.0, 1.0, 3.0)], dtype=wp.vec3).reshape((1, 1))
    vec = wp.array([wp.normalize(wp.vec3(0.1, 0.2, -1.0))], dtype=wp.vec3).reshape((1, 1))
    dist, geomid, normal = mjw.ray(m, d, pnt, vec)
    wp.synchronize()
    geomid_np = geomid.numpy()[0, 0]
    dist_np = dist.numpy()[0, 0]
    _assert_eq(geomid_np, 0, "geom_id")
    pnt_np, vec_np = pnt.numpy()[0, 0], vec.numpy()[0, 0]
    unused = np.zeros(1, dtype=np.int32)
    mj_dist = mujoco.mj_ray(mjm, mjd, pnt_np, vec_np, None, 1, -1, unused)
    _assert_eq(dist_np, mj_dist, "dist")

    # test that bvh raycast produces the same results
    rc = mjw.create_render_context(mjm)
    dist, geomid, normal = mjw.ray(m, d, pnt, vec, rc=rc)
    wp.synchronize()
    bvh_geomid_np = geomid.numpy()[0, 0]
    bvh_dist_np = dist.numpy()[0, 0]
    _assert_eq(bvh_geomid_np, geomid_np, "geom_id")
    _assert_eq(bvh_dist_np, dist_np, "dist")

    # looking on wrong side of plane
    pnt = wp.array([wp.vec3(0.0, 0.0, -0.5)], dtype=wp.vec3).reshape((1, 1))
    dist, geomid, normal = mjw.ray(m, d, pnt, vec)
    wp.synchronize()
    geomid_np = geomid.numpy()[0, 0]
    dist_np = dist.numpy()[0, 0]
    _assert_eq(geomid_np, -1, "geom_id")
    _assert_eq(dist_np, -1, "dist")

    # test that bvh raycast produces the same results
    dist, geomid, normal = mjw.ray(m, d, pnt, vec, rc=rc)
    wp.synchronize()
    bvh_geomid_np = geomid.numpy()[0, 0]
    bvh_dist_np = dist.numpy()[0, 0]
    _assert_eq(bvh_geomid_np, geomid_np, "geom_id")
    _assert_eq(bvh_dist_np, dist_np, "dist")

  def test_ray_sphere(self):
    """Tests ray<>sphere matches MuJoCo."""
    mjm, mjd, m, d = test_data.fixture("ray.xml")

    # looking down at sphere at a slight angle
    pnt = wp.array([wp.vec3(0.0, 0.0, 1.6)], dtype=wp.vec3).reshape((1, 1))
    vec = wp.array([wp.normalize(wp.vec3(0.1, 0.2, -1.0))], dtype=wp.vec3).reshape((1, 1))
    dist, geomid, normal = mjw.ray(m, d, pnt, vec)
    wp.synchronize()
    geomid_np = geomid.numpy()[0, 0]
    dist_np = dist.numpy()[0, 0]
    _assert_eq(geomid_np, 1, "geom_id")
    pnt_np, vec_np = pnt.numpy()[0, 0], vec.numpy()[0, 0]
    unused = np.zeros(1, dtype=np.int32)
    mj_dist = mujoco.mj_ray(mjm, mjd, pnt_np, vec_np, None, 1, -1, unused)
    _assert_eq(dist_np, mj_dist, "dist")

    # test that bvh raycast produces the same results
    rc = mjw.create_render_context(mjm)
    dist, geomid, normal = mjw.ray(m, d, pnt, vec, rc=rc)
    wp.synchronize()
    bvh_geomid_np = geomid.numpy()[0, 0]
    bvh_dist_np = dist.numpy()[0, 0]
    _assert_eq(bvh_geomid_np, geomid_np, "geom_id")
    _assert_eq(bvh_dist_np, dist_np, "dist")

  def test_ray_capsule(self):
    """Tests ray<>capsule matches MuJoCo."""
    mjm, mjd, m, d = test_data.fixture("ray.xml")

    # looking down at capsule at a slight angle
    pnt = wp.array([wp.vec3(0.5, 1.0, 1.6)], dtype=wp.vec3).reshape((1, 1))
    vec = wp.array([wp.normalize(wp.vec3(0.0, 0.05, -1.0))], dtype=wp.vec3).reshape((1, 1))
    dist, geomid, normal = mjw.ray(m, d, pnt, vec)
    wp.synchronize()
    geomid_np = geomid.numpy()[0, 0]
    dist_np = dist.numpy()[0, 0]
    _assert_eq(geomid_np, 2, "geom_id")
    pnt_np, vec_np = pnt.numpy()[0, 0], vec.numpy()[0, 0]
    unused = np.zeros(1, dtype=np.int32)
    mj_dist = mujoco.mj_ray(mjm, mjd, pnt_np, vec_np, None, 1, -1, unused)
    _assert_eq(dist_np, mj_dist, "dist")

    # test that bvh raycast produces the same results
    rc = mjw.create_render_context(mjm)
    dist, geomid, normal = mjw.ray(m, d, pnt, vec, rc=rc)
    wp.synchronize()
    bvh_geomid_np = geomid.numpy()[0, 0]
    bvh_dist_np = dist.numpy()[0, 0]
    _assert_eq(bvh_geomid_np, geomid_np, "geom_id")
    _assert_eq(bvh_dist_np, dist_np, "dist")

    # looking up at capsule from below
    pnt = wp.array([wp.vec3(-0.5, 1.0, 0.05)], dtype=wp.vec3).reshape((1, 1))
    vec = wp.array([wp.normalize(wp.vec3(0.0, 0.05, 1.0))], dtype=wp.vec3).reshape((1, 1))
    dist, geomid, normal = mjw.ray(m, d, pnt, vec)
    wp.synchronize()
    geomid_np = geomid.numpy()[0, 0]
    dist_np = dist.numpy()[0, 0]
    _assert_eq(geomid_np, 2, "geom_id")
    pnt_np, vec_np = pnt.numpy()[0, 0], vec.numpy()[0, 0]
    unused = np.zeros(1, dtype=np.int32)
    mj_dist = mujoco.mj_ray(mjm, mjd, pnt_np, vec_np, None, 1, -1, unused)
    _assert_eq(dist_np, mj_dist, "dist")

    # test that bvh raycast produces the same results
    dist, geomid, normal = mjw.ray(m, d, pnt, vec, rc=rc)
    wp.synchronize()
    bvh_geomid_np = geomid.numpy()[0, 0]
    bvh_dist_np = dist.numpy()[0, 0]
    _assert_eq(bvh_geomid_np, geomid_np, "geom_id")
    _assert_eq(bvh_dist_np, dist_np, "dist")

    # looking at cylinder of capsule from the side
    pnt = wp.array([wp.vec3(0.0, 1.0, 0.75)], dtype=wp.vec3).reshape((1, 1))
    vec = wp.array([wp.normalize(wp.vec3(1.0, 0.0, 0.0))], dtype=wp.vec3).reshape((1, 1))
    dist, geomid, normal = mjw.ray(m, d, pnt, vec)
    wp.synchronize()
    geomid_np = geomid.numpy()[0, 0]
    dist_np = dist.numpy()[0, 0]
    _assert_eq(geomid_np, 2, "geom_id")
    pnt_np, vec_np = pnt.numpy()[0, 0], vec.numpy()[0, 0]
    unused = np.zeros(1, dtype=np.int32)
    mj_dist = mujoco.mj_ray(mjm, mjd, pnt_np, vec_np, None, 1, -1, unused)
    _assert_eq(dist_np, mj_dist, "dist")

    # test that bvh raycast produces the same results
    dist, geomid, normal = mjw.ray(m, d, pnt, vec, rc=rc)
    wp.synchronize()
    bvh_geomid_np = geomid.numpy()[0, 0]
    bvh_dist_np = dist.numpy()[0, 0]
    _assert_eq(bvh_geomid_np, geomid_np, "geom_id")
    _assert_eq(bvh_dist_np, dist_np, "dist")

  def test_ray_cylinder(self):
    """Tests ray<>cylinder matches MuJoCo."""
    mjm, mjd, m, d = test_data.fixture("ray.xml")

    pnt = wp.array([wp.vec3(2.0, 0.0, 0.05)], dtype=wp.vec3).reshape((1, 1))
    vec = wp.array([wp.normalize(wp.vec3(0.0, 0.05, 1.0))], dtype=wp.vec3).reshape((1, 1))

    mj_geomid = np.zeros(1, dtype=np.int32)
    mj_dist = mujoco.mj_ray(mjm, mjd, pnt.numpy()[0, 0], vec.numpy()[0, 0], None, 1, -1, mj_geomid)
    dist, geomid, normal = mjw.ray(m, d, pnt, vec)
    wp.synchronize()
    geomid_np = geomid.numpy()[0, 0]
    dist_np = dist.numpy()[0, 0]

    _assert_eq(geomid_np, mj_geomid[0], "geomid")
    _assert_eq(dist_np, mj_dist, "dist")

    # test that bvh raycast produces the same results
    rc = mjw.create_render_context(mjm)
    dist, geomid, normal = mjw.ray(m, d, pnt, vec, rc=rc)
    wp.synchronize()
    bvh_geomid_np = geomid.numpy()[0, 0]
    bvh_dist_np = dist.numpy()[0, 0]
    _assert_eq(bvh_geomid_np, geomid_np, "geom_id")
    _assert_eq(bvh_dist_np, dist_np, "dist")

  def test_ray_box(self):
    """Tests ray<>box matches MuJoCo."""
    mjm, mjd, m, d = test_data.fixture("ray.xml")

    # looking down at box at a slight angle
    pnt = wp.array([wp.vec3(1.0, 0.0, 1.6)], dtype=wp.vec3).reshape((1, 1))
    vec = wp.array([wp.normalize(wp.vec3(0.0, 0.05, -1.0))], dtype=wp.vec3).reshape((1, 1))
    dist, geomid, normal = mjw.ray(m, d, pnt, vec)
    wp.synchronize()
    geomid_np = geomid.numpy()[0, 0]
    dist_np = dist.numpy()[0, 0]
    _assert_eq(geomid_np, 3, "geom_id")
    pnt_np, vec_np = pnt.numpy()[0, 0], vec.numpy()[0, 0]
    unused = np.zeros(1, dtype=np.int32)
    mj_dist = mujoco.mj_ray(mjm, mjd, pnt_np, vec_np, None, 1, -1, unused)
    _assert_eq(dist_np, mj_dist, "dist")

    # test that bvh raycast produces the same results
    rc = mjw.create_render_context(mjm)
    dist, geomid, normal = mjw.ray(m, d, pnt, vec, rc=rc)
    wp.synchronize()
    bvh_geomid_np = geomid.numpy()[0, 0]
    bvh_dist_np = dist.numpy()[0, 0]
    _assert_eq(bvh_geomid_np, geomid_np, "geom_id")
    _assert_eq(bvh_dist_np, dist_np, "dist")

    # looking up at box from below
    pnt = wp.array([wp.vec3(1.0, 0.0, 0.05)], dtype=wp.vec3).reshape((1, 1))
    vec = wp.array([wp.normalize(wp.vec3(0.0, 0.05, 1.0))], dtype=wp.vec3).reshape((1, 1))
    dist, geomid, normal = mjw.ray(m, d, pnt, vec)
    wp.synchronize()
    geomid_np = geomid.numpy()[0, 0]
    dist_np = dist.numpy()[0, 0]
    _assert_eq(geomid_np, 3, "geom_id")
    pnt_np, vec_np = pnt.numpy()[0, 0], vec.numpy()[0, 0]
    unused = np.zeros(1, dtype=np.int32)
    mj_dist = mujoco.mj_ray(mjm, mjd, pnt_np, vec_np, None, 1, -1, unused)
    _assert_eq(dist_np, mj_dist, "dist")

    # test that bvh raycast produces the same results
    dist, geomid, normal = mjw.ray(m, d, pnt, vec, rc=rc)
    wp.synchronize()
    bvh_geomid_np = geomid.numpy()[0, 0]
    bvh_dist_np = dist.numpy()[0, 0]
    _assert_eq(bvh_geomid_np, geomid_np, "bvh geom_id")
    _assert_eq(bvh_dist_np, dist_np, "bvh dist")

  def test_ray_mesh(self):
    """Tests ray<>mesh matches MuJoCo."""
    mjm, mjd, m, d = test_data.fixture("ray.xml")

    # look at the tetrahedron
    pnt = wp.array([wp.vec3(2.0, 2.0, 2.0)], dtype=wp.vec3).reshape((1, 1))
    vec = wp.array([wp.normalize(wp.vec3(-1.0, -1.0, -1.0))], dtype=wp.vec3).reshape((1, 1))
    dist, geomid, normal = mjw.ray(m, d, pnt, vec)
    wp.synchronize()
    geomid_np = geomid.numpy()[0, 0]
    dist_np = dist.numpy()[0, 0]
    _assert_eq(geomid_np, 4, "geom_id")

    pnt_np, vec_np = pnt.numpy()[0, 0], vec.numpy()[0, 0]
    unused = np.zeros(1, dtype=np.int32)
    mj_dist = mujoco.mj_ray(mjm, mjd, pnt_np, vec_np, None, 1, -1, unused)
    _assert_eq(dist_np, mj_dist, "dist-tetrahedron")

    # test that bvh raycast produces the same results
    rc = mjw.create_render_context(mjm)
    dist, geomid, normal = mjw.ray(m, d, pnt, vec, rc=rc)
    wp.synchronize()
    bvh_geomid_np = geomid.numpy()[0, 0]
    bvh_dist_np = dist.numpy()[0, 0]
    _assert_eq(bvh_geomid_np, geomid_np, "geom_id")
    _assert_eq(bvh_dist_np, dist_np, "dist")

    # look away from the dodecahedron
    pnt = wp.array([wp.vec3(4.0, 2.0, 2.0)], dtype=wp.vec3).reshape((1, 1))
    vec = wp.array([wp.normalize(wp.vec3(2.0, 1.0, 1.0))], dtype=wp.vec3).reshape((1, 1))
    dist, geomid, normal = mjw.ray(m, d, pnt, vec)
    wp.synchronize()
    geomid_np = geomid.numpy()[0, 0]
    _assert_eq(geomid_np, -1, "geom_id")

    # test that bvh raycast produces the same results
    dist, geomid, normal = mjw.ray(m, d, pnt, vec, rc=rc)
    wp.synchronize()
    bvh_geomid_np = geomid.numpy()[0, 0]
    _assert_eq(bvh_geomid_np, geomid_np, "geom_id")

    # look at the dodecahedron
    pnt = wp.array([wp.vec3(4.0, 2.0, 2.0)], dtype=wp.vec3).reshape((1, 1))
    vec = wp.array([wp.normalize(wp.vec3(-2.0, -1.0, -1.0))], dtype=wp.vec3).reshape((1, 1))
    dist, geomid, normal = mjw.ray(m, d, pnt, vec)
    wp.synchronize()
    geomid_np = geomid.numpy()[0, 0]
    dist_np = dist.numpy()[0, 0]
    _assert_eq(geomid_np, 5, "geom_id")

    pnt_np, vec_np = pnt.numpy()[0, 0], vec.numpy()[0, 0]
    unused = np.zeros(1, dtype=np.int32)
    mj_dist = mujoco.mj_ray(mjm, mjd, pnt_np, vec_np, None, 1, -1, unused)
    _assert_eq(dist_np, mj_dist, "dist-dodecahedron")

    # test that bvh raycast produces the same results
    dist, geomid, normal = mjw.ray(m, d, pnt, vec, rc=rc)
    wp.synchronize()
    bvh_geomid_np = geomid.numpy()[0, 0]
    bvh_dist_np = dist.numpy()[0, 0]
    _assert_eq(bvh_geomid_np, geomid_np, "bvh geom_id")
    _assert_eq(bvh_dist_np, dist_np, "bvh dist")

  def test_ray_hfield(self):
    mjm, mjd, m, d = test_data.fixture("ray.xml")

    pnt = wp.array([wp.vec3(0.0, 2.0, 2.0)], dtype=wp.vec3).reshape((1, 1))
    vec = wp.array([wp.vec3(0.0, 0.0, -1.0)], dtype=wp.vec3).reshape((1, 1))
    dist, geomid, normal = mjw.ray(m, d, pnt, vec)

    mj_geomid = np.zeros(1, dtype=np.int32)
    mj_dist = mujoco.mj_ray(mjm, mjd, pnt.numpy()[0, 0], vec.numpy()[0, 0], None, 1, -1, mj_geomid)
    wp.synchronize()
    geomid_np = geomid.numpy()[0, 0]
    dist_np = dist.numpy()[0, 0]

    _assert_eq(dist_np, mj_dist, "dist")
    _assert_eq(geomid_np, mj_geomid[0], "geomid")

    # test that bvh raycast produces the same results
    rc = mjw.create_render_context(mjm)
    dist, geomid, normal = mjw.ray(m, d, pnt, vec, rc=rc)
    wp.synchronize()
    bvh_geomid_np = geomid.numpy()[0, 0]
    bvh_dist_np = dist.numpy()[0, 0]
    _assert_eq(bvh_geomid_np, geomid_np, "geom_id")
    _assert_eq(bvh_dist_np, dist_np, "bvh dist")

  def test_ray_geomgroup(self):
    """Tests ray geomgroup filter."""
    mjm, mjd, m, d = test_data.fixture("ray.xml")

    # hits plane with geom_group[0] = 1
    pnt = wp.array([wp.vec3(2.0, 1.0, 3.0)], dtype=wp.vec3).reshape((1, 1))
    vec = wp.array([wp.normalize(wp.vec3(0.1, 0.2, -1.0))], dtype=wp.vec3).reshape((1, 1))
    geomgroup = vec6(1, 0, 0, 0, 0, 0)
    dist, geomid, normal = mjw.ray(m, d, pnt, vec, geomgroup=geomgroup)
    wp.synchronize()
    geomid_np = geomid.numpy()[0, 0]
    dist_np = dist.numpy()[0, 0]
    _assert_eq(geomid_np, 0, "geom_id")

    pnt_np, vec_np = pnt.numpy()[0, 0], vec.numpy()[0, 0]
    unused = np.zeros(1, dtype=np.int32)
    mj_dist = mujoco.mj_ray(mjm, mjd, pnt_np, vec_np, None, 1, -1, unused)
    _assert_eq(dist_np, mj_dist, "dist")

    # test that bvh raycast produces the same results
    rc = mjw.create_render_context(mjm)
    dist, geomid, normal = mjw.ray(m, d, pnt, vec, geomgroup=geomgroup, rc=rc)
    wp.synchronize()
    bvh_geomid_np = geomid.numpy()[0, 0]
    bvh_dist_np = dist.numpy()[0, 0]
    _assert_eq(bvh_geomid_np, geomid_np, "geom_id")
    _assert_eq(bvh_dist_np, dist_np, "dist")

    # nothing hit with geom_group[0] = 0
    pnt = wp.array([wp.vec3(2.0, 1.0, 3.0)], dtype=wp.vec3).reshape((1, 1))
    vec = wp.array([wp.normalize(wp.vec3(0.1, 0.2, -1.0))], dtype=wp.vec3).reshape((1, 1))
    geomgroup = vec6(0, 0, 0, 0, 0, 0)
    dist, geomid, normal = mjw.ray(m, d, pnt, vec, geomgroup=geomgroup)
    wp.synchronize()
    geomid_np = geomid.numpy()[0, 0]
    dist_np = dist.numpy()[0, 0]
    _assert_eq(geomid_np, -1, "geom_id")
    _assert_eq(dist_np, -1, "dist")

    # test that bvh raycast produces the same results
    dist, geomid, normal = mjw.ray(m, d, pnt, vec, geomgroup=geomgroup, rc=rc)
    wp.synchronize()
    bvh_geomid_np = geomid.numpy()[0, 0]
    bvh_dist_np = dist.numpy()[0, 0]
    _assert_eq(bvh_geomid_np, geomid_np, "geom_id")
    _assert_eq(bvh_dist_np, dist_np, "dist")

  def test_ray_flg_static(self):
    """Tests ray flg_static filter."""
    mjm, mjd, m, d = test_data.fixture("ray.xml")

    # nothing hit with flg_static = False
    pnt = wp.array([wp.vec3(2.0, 1.0, 3.0)], dtype=wp.vec3).reshape((1, 1))
    vec = wp.array([wp.normalize(wp.vec3(0.1, 0.2, -1.0))], dtype=wp.vec3).reshape((1, 1))
    dist, geomid, normal = mjw.ray(m, d, pnt, vec, flg_static=False)
    wp.synchronize()
    geomid_np = geomid.numpy()[0, 0]
    dist_np = dist.numpy()[0, 0]
    _assert_eq(geomid_np, -1, "geom_id")
    _assert_eq(dist_np, -1, "dist")

    # test that bvh raycast produces the same results
    rc = mjw.create_render_context(mjm)
    dist, geomid, normal = mjw.ray(m, d, pnt, vec, flg_static=False, rc=rc)
    wp.synchronize()
    bvh_geomid_np = geomid.numpy()[0, 0]
    bvh_dist_np = dist.numpy()[0, 0]
    _assert_eq(bvh_geomid_np, geomid_np, "geom_id")
    _assert_eq(bvh_dist_np, dist_np, "dist")

  def test_ray_bodyexclude(self):
    """Tests ray bodyexclude filter."""
    mjm, mjd, m, d = test_data.fixture("ray.xml")

    # nothing hit with bodyexclude = 0 (world body)
    pnt = wp.array([wp.vec3(2.0, 1.0, 3.0)], dtype=wp.vec3).reshape((1, 1))
    vec = wp.array([wp.normalize(wp.vec3(0.1, 0.2, -1.0))], dtype=wp.vec3).reshape((1, 1))
    dist, geomid, normal = mjw.ray(m, d, pnt, vec, bodyexclude=0)
    wp.synchronize()
    geomid_np = geomid.numpy()[0, 0]
    dist_np = dist.numpy()[0, 0]
    normal_np = normal.numpy()[0, 0]
    _assert_eq(geomid_np, -1, "geom_id")
    _assert_eq(dist_np, -1, "dist")
    _assert_eq(normal_np, 0, "normal")

    # test that bvh raycast produces the same results
    rc = mjw.create_render_context(mjm)
    dist, geomid, normal = mjw.ray(m, d, pnt, vec, bodyexclude=0, rc=rc)
    wp.synchronize()
    bvh_geomid_np = geomid.numpy()[0, 0]
    bvh_dist_np = dist.numpy()[0, 0]
    bvh_normal_np = normal.numpy()[0, 0]
    _assert_eq(bvh_geomid_np, geomid_np, "geom_id")
    _assert_eq(bvh_dist_np, dist_np, "dist")
    _assert_eq(bvh_normal_np, normal_np, "normal")

  def test_ray_invisible(self):
    """Tests ray doesn't hit transparent geoms."""
    mjm, mjd, m, d = test_data.fixture("ray.xml")

    # nothing hit with transparent geoms
    m.geom_rgba = wp.array2d([[wp.vec4(0.0, 0.0, 0.0, 0.0)] * 8], dtype=wp.vec4)
    mujoco.mj_forward(mjm, mjd)

    pnt = wp.array([wp.vec3(2.0, 1.0, 3.0)], dtype=wp.vec3).reshape((1, 1))
    vec = wp.array([wp.normalize(wp.vec3(0.1, 0.2, -1.0))], dtype=wp.vec3).reshape((1, 1))
    dist, geomid, normal = mjw.ray(m, d, pnt, vec)
    wp.synchronize()
    geomid_np = geomid.numpy()[0, 0]
    dist_np = dist.numpy()[0, 0]
    normal_np = normal.numpy()[0, 0]
    _assert_eq(geomid_np, -1, "geom_id")
    _assert_eq(dist_np, -1, "dist")
    _assert_eq(normal_np, 0, "normal")

    # test that bvh raycast produces the same results
    rc = mjw.create_render_context(mjm)
    dist, geomid, normal = mjw.ray(m, d, pnt, vec, rc=rc)
    wp.synchronize()
    bvh_geomid_np = geomid.numpy()[0, 0]
    bvh_dist_np = dist.numpy()[0, 0]
    bvh_normal_np = normal.numpy()[0, 0]
    _assert_eq(bvh_geomid_np, geomid_np, "geom_id")
    _assert_eq(bvh_dist_np, dist_np, "dist")
    _assert_eq(bvh_normal_np, normal_np, "normal")


if __name__ == "__main__":
  absltest.main()



import warp as wp


@wp.func
def Fract(x: float) -> float:
  return x - wp.floor(x)


@wp.func
def Subtraction(a: float, b: float) -> float:
  return wp.max(a, -b)


@wp.func
def Union(a: float, b: float) -> float:
  return wp.min(a, b)


@wp.func
def Intersection(a: float, b: float) -> float:
  return wp.max(a, b)


@wp.func
def bolt(p: wp.vec3, attr: wp.vec3) -> float:
  screw = 12.0
  radius = wp.sqrt(p[0] * p[0] + p[1] * p[1]) - attr[0]
  sqrt12 = wp.sqrt(2.0) / 2.0

  azimuth = wp.atan2(p[1], p[0])
  triangle = wp.abs(Fract(p[2] * screw - azimuth / wp.pi / 2.0) - 0.5)
  thread = (radius - triangle / screw) * sqrt12

  bolt_val = Subtraction(thread, 0.5 - wp.abs(p[2] + 0.5))
  cone = (p[2] - radius) * sqrt12

  bolt_val = Subtraction(bolt_val, cone + 1.0 * sqrt12)

  point2D = wp.vec2(p[0], p[1])
  k = 6.0 / wp.pi / 2.0
  angle = -wp.floor((wp.atan2(point2D[1], point2D[0])) * k + 0.5) / k
  s = wp.vec2(wp.sin(angle), wp.sin(angle + wp.pi * 0.5))

  res = wp.vec2(s[1] * point2D[0] - s[0] * point2D[1], s[0] * point2D[0] + s[1] * point2D[1])
  point3D = wp.vec3(res[0], res[1], p[2])
  head = point3D[0] - 0.5

  head = Intersection(head, wp.abs(point3D[2] + 0.25) - 0.25)
  head = Intersection(head, (point3D[2] + radius - 0.22) * sqrt12)

  return Union(bolt_val, head)


@wp.func
def bolt_sdf_grad(p: wp.vec3, attr: wp.vec3) -> wp.vec3:
  grad = wp.vec3()
  eps = 1e-6
  f_original = bolt(p, attr)
  x_plus = wp.vec3(p[0] + eps, p[1], p[2])
  f_plus = bolt(x_plus, attr)
  grad[0] = (f_plus - f_original) / eps

  x_plus = wp.vec3(p[0], p[1] + eps, p[2])
  f_plus = bolt(x_plus, attr)
  grad[1] = (f_plus - f_original) / eps

  x_plus = wp.vec3(p[0], p[1], p[2] + eps)
  f_plus = bolt(x_plus, attr)
  grad[2] = (f_plus - f_original) / eps
  return grad



# Copyright 2025 The Newton Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Tests for support functions."""

import mujoco
import numpy as np
import warp as wp
from absl.testing import absltest
from absl.testing import parameterized

import mujoco_warp as mjwarp
from mujoco_warp import ConeType
from mujoco_warp import State
from mujoco_warp import test_data
from mujoco_warp._src.block_cholesky import create_blocked_cholesky_func
from mujoco_warp._src.block_cholesky import create_blocked_cholesky_solve_func

# tolerance for difference between MuJoCo and MJWarp support calculations - mostly
# due to float precision
_TOLERANCE = 5e-5


def _assert_eq(a, b, name):
  tol = _TOLERANCE * 10  # avoid test noise
  err_msg = f"mismatch: {name}"
  np.testing.assert_allclose(a, b, err_msg=err_msg, atol=tol, rtol=tol)


class SupportTest(parameterized.TestCase):
  @parameterized.parameters(mujoco.mjtJacobian.mjJAC_SPARSE, mujoco.mjtJacobian.mjJAC_DENSE)
  def test_mul_m(self, jacobian):
    """Tests mul_m."""
    mjm, mjd, m, d = test_data.fixture("pendula.xml", overrides={"opt.jacobian": jacobian})

    mj_res = np.zeros(mjm.nv)
    mj_vec = np.random.uniform(low=-1.0, high=1.0, size=mjm.nv)
    mujoco.mj_mulM(mjm, mjd, mj_res, mj_vec)

    res = wp.zeros((1, mjm.nv), dtype=wp.float32)
    vec = wp.from_numpy(np.expand_dims(mj_vec, axis=0), dtype=wp.float32)
    mjwarp.mul_m(m, d, res, vec)

    _assert_eq(res.numpy()[0], mj_res, f"mul_m ({jacobian})")

  def test_xfrc_accumulated(self):
    """Tests that xfrc_accumulate output matches mj_xfrcAccumulate."""
    mjm, mjd, m, d = test_data.fixture("pendula.xml")
    xfrc = np.random.randn(*d.xfrc_applied.numpy().shape)
    d.xfrc_applied = wp.from_numpy(xfrc, dtype=wp.spatial_vector)
    qfrc = wp.zeros((1, mjm.nv), dtype=wp.float32)
    mjwarp.xfrc_accumulate(m, d, qfrc)

    qfrc_expected = np.zeros(m.nv)
    xfrc = xfrc[0]
    for i in range(1, m.nbody):
      mujoco.mj_applyFT(mjm, mjd, xfrc[i, :3], xfrc[i, 3:], mjd.xipos[i], i, qfrc_expected)
    np.testing.assert_almost_equal(qfrc.numpy()[0], qfrc_expected, 6)

  @parameterized.parameters(
    (ConeType.PYRAMIDAL, 1, False),
    (ConeType.PYRAMIDAL, 3, False),
    (ConeType.PYRAMIDAL, 4, False),
    (ConeType.PYRAMIDAL, 6, False),
    (ConeType.PYRAMIDAL, 1, True),
    (ConeType.PYRAMIDAL, 3, True),
    (ConeType.PYRAMIDAL, 4, True),
    (ConeType.PYRAMIDAL, 6, True),
    (ConeType.ELLIPTIC, 1, False),
    (ConeType.ELLIPTIC, 3, False),
    (ConeType.ELLIPTIC, 4, False),
    (ConeType.ELLIPTIC, 6, False),
    (ConeType.ELLIPTIC, 1, True),
    (ConeType.ELLIPTIC, 3, True),
    (ConeType.ELLIPTIC, 4, True),
    (ConeType.ELLIPTIC, 6, True),
  )
  def test_contact_force(self, cone, condim, to_world_frame):
    _CONTACT = f"""
      <mujoco>
        <worldbody>
          <geom type="plane" size="10 10 .001"/>
          <body pos="0 0 1">
            <freejoint/>
            <geom fromto="-.4 0 0 .4 0 0" size=".05 .1" type="capsule" condim="{condim}" friction="1 1 1"/>
          </body>
        </worldbody>
        <keyframe>
          <key qpos="0 0 0.04 1 0 0 0" qvel="-1 -1 -1 .1 .1 .1"/>
        </keyframe>
      </mujoco>
    """
    mjm, mjd, m, d = test_data.fixture(xml=_CONTACT, keyframe=0, overrides={"opt.cone": cone})

    mj_force = np.zeros(6, dtype=float)
    mujoco.mj_contactForce(mjm, mjd, 0, mj_force)

    contact_ids = wp.zeros(1, dtype=int)
    force = wp.zeros(1, dtype=wp.spatial_vector)

    mjwarp.contact_force(m, d, contact_ids, to_world_frame, force)

    if to_world_frame:
      frame = mjd.contact.frame[0].reshape((3, 3))
      mj_force = np.concatenate([frame.T @ mj_force[:3], frame.T @ mj_force[3:]])

    _assert_eq(force.numpy()[0], mj_force, "contact force")

  def test_get_state(self):
    mjm, mjd, m, d = test_data.fixture(
      "constraints.xml", keyframe=0, ctrl_noise=1.0, qfrc_noise=1.0, xfrc_noise=1.0, mocap_noise=1.0
    )

    size = mujoco.mj_stateSize(mjm, mujoco.mjtState.mjSTATE_INTEGRATION)

    mj_state = np.zeros(size)
    mujoco.mj_getState(mjm, mjd, mj_state, mujoco.mjtState.mjSTATE_INTEGRATION)

    mjw_state = wp.zeros((d.nworld, size), dtype=float)
    mjwarp.get_state(m, d, mjw_state, State.INTEGRATION)

    _assert_eq(mjw_state.numpy()[0], mj_state, "state")

    d2 = mjwarp.put_data(mjm, mjd, nworld=2)
    mjw_state2 = wp.zeros((d2.nworld, size), dtype=float)
    active = wp.array([False, True], dtype=bool)
    mjwarp.get_state(m, d2, mjw_state2, State.INTEGRATION, active=active)

    _assert_eq(mjw_state2.numpy()[0], 0, "state0")
    _assert_eq(mjw_state2.numpy()[1], mj_state, "state1")

  def test_set_state(self):
    mjm, mjd, m, d = test_data.fixture(
      "constraints.xml", keyframe=0, ctrl_noise=1.0, qfrc_noise=1.0, xfrc_noise=1.0, mocap_noise=1.0
    )

    size = mujoco.mj_stateSize(mjm, mujoco.mjtState.mjSTATE_INTEGRATION)

    mj_state = np.zeros(size)
    mujoco.mj_getState(mjm, mjd, mj_state, mujoco.mjtState.mjSTATE_INTEGRATION)

    mjw_state = wp.array(np.expand_dims(mj_state, axis=0), dtype=float)
    mjwarp.set_state(m, d, mjw_state, State.INTEGRATION)

    state_integration = [
      "time",
      "qpos",
      "qvel",
      "act",
      "qacc_warmstart",
      "ctrl",
      "qfrc_applied",
      "xfrc_applied",
      "eq_active",
      "mocap_pos",
      "mocap_quat",
    ]

    for field in state_integration:
      _assert_eq(getattr(d, field).numpy()[0], getattr(mjd, field), field)

    d2 = mjwarp.make_data(mjm, nworld=2)
    mjw_state2 = wp.array(np.tile(mj_state, (2, 1)), dtype=float)
    active = wp.array([False, True], dtype=bool)
    mjwarp.set_state(m, d2, mjw_state2, State.INTEGRATION, active=active)

    for field in state_integration:
      _assert_eq(getattr(d2, field).numpy()[1], getattr(mjd, field), field)

  def test_block_cholesky(self):
    """Tests block Cholesky decomposition and solve against numpy using n_humanoid model."""
    mjm, mjd, m, d = test_data.fixture("humanoid/n_humanoid.xml")

    # Add noise and initialize
    np.random.seed(42)
    mjd.qpos[:] += np.random.uniform(-0.1, 0.1, mjm.nq)
    mjd.qvel[:] += np.random.uniform(-0.1, 0.1, mjm.nv)
    mujoco.mj_step(mjm, mjd, 10)
    mujoco.mj_forward(mjm, mjd)

    m = mjwarp.put_model(mjm)
    d = mjwarp.put_data(mjm, mjd, nworld=1)

    # Run forward to populate everything including efc.h
    mjwarp.forward(m, d)

    # Get the constraint Hessian matrix size
    nv = m.nv
    nv_pad = m.nv_pad
    nworld = d.nworld

    # Create combined factor and solve kernel as in solver.py
    @wp.kernel(module="unique", enable_backward=False)
    def combined_cholesky_kernel(
      grad_in: wp.array3d(dtype=float),
      h_in: wp.array3d(dtype=float),
      done_in: wp.array(dtype=bool),
      hfactor_in: wp.array3d(dtype=float),
      Mgrad_out: wp.array3d(dtype=float),
    ):
      worldid = wp.tid()
      TILE_SIZE = wp.static(16)

      if done_in[worldid]:
        return

      wp.static(create_blocked_cholesky_func(TILE_SIZE))(h_in[worldid], nv_pad, hfactor_in[worldid])
      wp.static(create_blocked_cholesky_solve_func(TILE_SIZE, nv_pad))(
        hfactor_in[worldid], grad_in[worldid], nv_pad, Mgrad_out[worldid]
      )

    # Create test vector and fill the built-in arrays
    b = np.random.randn(nv).astype(np.float32)

    # 1. Generate a random SPD matrix for active region
    A = np.random.randn(nv, nv).astype(np.float32)
    SPD_active_hessian = A @ A.T + nv * np.eye(nv, dtype=A.dtype)  # Make symmetric & strongly PD

    # 2. Copy the active SPD region into h_np[0, :nv, :nv]
    h_np = np.zeros((nworld, nv_pad, nv_pad), dtype=float)
    h_np[0, :nv, :nv] = SPD_active_hessian

    # Add identity to the padding region
    padding_size = nv_pad - nv
    if padding_size > 0:
      h_np[0, nv:, nv:] = np.eye(padding_size, dtype=np.float32)
    h = wp.array(h_np, dtype=float)

    # 3. Create inline arrays for grad, Mgrad, and done (no longer in d.efc)
    grad_np = np.zeros((nworld, nv_pad))
    grad_np[0, :nv] = b
    grad = wp.array(grad_np, dtype=float)

    # Zero out the temporary arrays to ensure clean state
    L_init = np.zeros((nworld, nv_pad, nv_pad), dtype=np.float32)
    # Initialize padding region to identity
    L_init[0, nv:, nv:] = np.eye(nv_pad - nv, dtype=np.float32)

    hfactor = wp.array(L_init, dtype=float)

    Mgrad = wp.zeros((nworld, nv_pad), dtype=float)

    # Ensure done is False so kernel executes
    done = wp.zeros((nworld,), dtype=bool)

    # Launch with same dimensions as solver.py, using inline arrays
    wp.launch_tiled(
      combined_cholesky_kernel,
      dim=nworld,
      inputs=[grad.reshape(shape=(nworld, nv_pad, 1)), h, done, hfactor],
      outputs=[Mgrad.reshape(shape=(nworld, nv_pad, 1))],
      block_dim=m.block_dim.update_gradient_cholesky,
    )
    wp.synchronize()

    L_result = hfactor.numpy()[0]
    x_result = Mgrad.numpy()[0]

    # Verify padding outside active region doesn't affect active computation
    # Off-diagonal padding should be zero (active region shouldn't touch padding)
    np.testing.assert_array_equal(
      L_result[nv:, :nv],
      0.0,
      err_msg="padding rows in active region were overwritten",
    )
    np.testing.assert_array_equal(
      L_result[:nv, nv:],
      0.0,
      err_msg="padding columns in active region were overwritten",
    )

    # Check that padding region remains identity after factorization
    padding_size = nv_pad - nv
    if padding_size > 0:
      padding_square = L_result[nv:, nv:]
      expected_identity = np.eye(padding_size, dtype=np.float32)
      np.testing.assert_allclose(
        padding_square,
        expected_identity,
        rtol=1e-6,
        atol=1e-6,
        err_msg="Padding region should remain identity after factorization",
      )

    # Compare with numpy cholesky on symmetrized Hessian
    L_numpy = np.linalg.cholesky(SPD_active_hessian)
    np.testing.assert_allclose(
      L_result[:nv, :nv],
      L_numpy,
      rtol=1e-4,
      atol=1e-4,
      err_msg="Cholesky decomposition mismatch with numpy",
    )

    # Verify L @ L.T = A (symmetrized Hessian)
    A_reconstructed = L_result[:nv, :nv] @ L_result[:nv, :nv].T
    np.testing.assert_allclose(
      A_reconstructed,
      SPD_active_hessian,
      rtol=1e-5,
      atol=1e-5,
      err_msg="L @ L.T does not equal symmetrized Hessian",
    )

    # Verify solution: A @ x = b using the symmetrized Hessian
    x_numpy = np.linalg.solve(SPD_active_hessian, b)
    np.testing.assert_allclose(
      x_result[:nv],
      x_numpy,
      rtol=1e-3,
      atol=1e-3,
      err_msg="Solution mismatch with numpy",
    )

  @parameterized.parameters(
    ("pendula.xml", 1),
    ("pendula.xml", 2),
    ("humanoid/n_humanoid.xml", 1),
    ("humanoid/n_humanoid.xml", 5),
    ("constraints.xml", 1),
  )
  def test_jac(self, xml, bodyid):
    """Tests jac against mj_jac."""
    mjm, mjd, m, d = test_data.fixture(xml)

    point = mjd.xipos[bodyid]

    jacp_mj = np.zeros((3, mjm.nv))
    jacr_mj = np.zeros((3, mjm.nv))
    mujoco.mj_jac(mjm, mjd, jacp_mj, jacr_mj, point, bodyid)

    point_wp = wp.array([point], dtype=wp.vec3)
    bodyid_wp = wp.array([bodyid], dtype=int)
    jacp_wp = wp.zeros((1, 3, mjm.nv), dtype=float)
    jacr_wp = wp.zeros((1, 3, mjm.nv), dtype=float)

    mjwarp.jac(m, d, jacp_wp, jacr_wp, point_wp, bodyid_wp)

    _assert_eq(jacp_wp.numpy()[0], jacp_mj, f"jacp ({xml}, body {bodyid})")
    _assert_eq(jacr_wp.numpy()[0], jacr_mj, f"jacr ({xml}, body {bodyid})")

  def test_jac_optional_outputs(self):
    """Tests jac with None outputs."""
    mjm, mjd, m, d = test_data.fixture("pendula.xml")

    point = mjd.xipos[1]
    point_wp = wp.array([point], dtype=wp.vec3)
    bodyid_wp = wp.array([1], dtype=int)

    jacp_wp = wp.zeros((1, 3, mjm.nv), dtype=float)
    jacr_wp = wp.zeros((1, 3, mjm.nv), dtype=float)

    jacp_mj = np.zeros((3, mjm.nv))
    jacr_mj = np.zeros((3, mjm.nv))
    mujoco.mj_jac(mjm, mjd, jacp_mj, jacr_mj, point, 1)

    mjwarp.jac(m, d, jacp_wp, None, point_wp, bodyid_wp)
    _assert_eq(jacp_wp.numpy()[0], jacp_mj, "jacp (optional jacr)")

    mjwarp.jac(m, d, None, jacr_wp, point_wp, bodyid_wp)
    _assert_eq(jacr_wp.numpy()[0], jacr_mj, "jacr (optional jacp)")

    mjwarp.jac(m, d, None, None, point_wp, bodyid_wp)

  def test_jac_nworld(self):
    """Tests jac with multiple worlds."""
    mjm, mjd, m, d = test_data.fixture("pendula.xml", nworld=2)

    bodyid = 1
    point = mjd.xipos[bodyid]

    jacp_mj = np.zeros((3, mjm.nv))
    jacr_mj = np.zeros((3, mjm.nv))
    mujoco.mj_jac(mjm, mjd, jacp_mj, jacr_mj, point, bodyid)

    point_wp = wp.array([point, point], dtype=wp.vec3)
    bodyid_wp = wp.array([bodyid, bodyid], dtype=int)
    jacp_wp = wp.zeros((2, 3, mjm.nv), dtype=float)
    jacr_wp = wp.zeros((2, 3, mjm.nv), dtype=float)

    mjwarp.jac(m, d, jacp_wp, jacr_wp, point_wp, bodyid_wp)

    for w in range(2):
      _assert_eq(jacp_wp.numpy()[w], jacp_mj, f"jacp world {w}")
      _assert_eq(jacr_wp.numpy()[w], jacr_mj, f"jacr world {w}")


if __name__ == "__main__":
  wp.init()
  absltest.main()



# Copyright 2025 The Newton Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Tests for solver functions."""

import mujoco
import numpy as np
import warp as wp
from absl.testing import absltest
from absl.testing import parameterized

import mujoco_warp as mjw
from mujoco_warp import ConeType
from mujoco_warp import SolverType
from mujoco_warp import test_data
from mujoco_warp._src import solver

# tolerance for difference between MuJoCo and MJWarp solver calculations - mostly
# due to float precision
_TOLERANCE = 5e-3


def _assert_eq(a, b, name):
  tol = _TOLERANCE * 20  # avoid test noise
  err_msg = f"mismatch: {name}"
  np.testing.assert_allclose(a, b, err_msg=err_msg, atol=tol, rtol=tol)


class SolverTest(parameterized.TestCase):
  @parameterized.product(cone=tuple(ConeType), solver_=tuple(SolverType))
  def test_constraint_update(self, cone, solver_):
    """Tests _update_constraint function is correct."""
    for keyframe in range(3):
      mjm, mjd, m, d = test_data.fixture(
        "constraints.xml",
        keyframe=keyframe,
        overrides={"opt.solver": solver_, "opt.cone": cone, "opt.iterations": 0},
      )

      def cost(qacc):
        jaref = np.zeros(mjd.nefc, dtype=float)
        cost = np.zeros(1)
        mujoco.mj_mulJacVec(mjm, mjd, jaref, qacc)
        mujoco.mj_constraintUpdate(mjm, mjd, jaref - mjd.efc_aref, cost, 0)
        return cost

      mjd_cost = cost(mjd.qacc)

      # solve with 0 iterations just initializes constraints and costs and then exits
      d.efc.force.zero_()
      d.qfrc_constraint.zero_()
      ctx = solver.create_solver_context(m, d)
      solver._solve(m, d, ctx)

      # Get the ordering indices based on efc_force, efc_state for MJWarp
      nefc = d.nefc.numpy()[0]
      efc_force = d.efc.force.numpy()[0, :nefc]
      efc_state = d.efc.state.numpy()[0, :nefc]
      # Get the ordering indices based on efc_force, efc_state for MuJoCo
      mjd_efc_force = mjd.efc_force[:nefc]
      mjd_efc_state = mjd.efc_state[:nefc]

      # Create sorting keys using lexsort (more efficient for multiple keys)
      d_sort_indices = np.lexsort((efc_force, efc_state))
      mjd_sort_indices = np.lexsort((mjd_efc_force, mjd_efc_state))

      solver.init_context(m, d, ctx, grad=False)
      ctx_cost = ctx.cost.numpy()[0] - ctx.gauss.numpy()[0]
      qfrc_constraint = d.qfrc_constraint.numpy()[0]

      efc_sorted_force = efc_force[d_sort_indices]
      efc_sorted_state = efc_state[d_sort_indices]
      mjd_sorted_force = mjd_efc_force[mjd_sort_indices]
      mjd_sorted_state = mjd_efc_state[mjd_sort_indices]

      _assert_eq(efc_sorted_state, mjd_sorted_state, "efc_state")
      _assert_eq(efc_sorted_force, mjd_sorted_force, "efc_force")
      _assert_eq(ctx_cost, mjd_cost, "cost")
      _assert_eq(qfrc_constraint, mjd.qfrc_constraint, "qfrc_constraint")

  @parameterized.product(ls_parallel=(True, False), cone=(ConeType.PYRAMIDAL, ConeType.ELLIPTIC))
  def test_init_linesearch(self, ls_parallel, cone):
    """Test linesearch initialization.

    Parallel linesearch has separate prep kernels that write quad, quad_gauss, jv.
    Iterative linesearch fuses these in-kernel: quad_gauss is internal, quad is
    only written for elliptic cones.
    """
    for keyframe in range(3):
      mjm, mjd, m, d = test_data.fixture(
        "constraints.xml",
        keyframe=keyframe,
        overrides={
          "opt.iterations": 0,
          "opt.ls_iterations": 1,
          "opt.ls_parallel": ls_parallel,
          "opt.cone": cone,
        },
      )

      # One step to obtain more non-zeros results
      mjw.step(m, d)

      # Create a SolverContext to access internal solver arrays
      ctx = solver.create_solver_context(m, d)
      solver._solve(m, d, ctx)

      # Calculate target values
      nefc = d.nefc.numpy()[0]
      ctx_search_np = ctx.search.numpy()[0]
      efc_J_np = d.efc.J.numpy()[0][:nefc, : m.nv]
      ctx_gauss_np = ctx.gauss.numpy()[0]
      efc_Ma_np = d.efc.Ma.numpy()[0]
      ctx_Jaref_np = ctx.Jaref.numpy()[0][:nefc]
      efc_D_np = d.efc.D.numpy()[0][:nefc]
      qfrc_smooth_np = d.qfrc_smooth.numpy()[0]

      target_mv = np.zeros(mjm.nv)
      mujoco.mj_mulM(mjm, mjd, target_mv, ctx_search_np)
      target_jv = efc_J_np @ ctx_search_np
      target_quad_gauss = np.array(
        [
          ctx_gauss_np,
          np.dot(ctx_search_np, efc_Ma_np - qfrc_smooth_np),
          0.5 * np.dot(ctx_search_np, target_mv),
        ]
      )
      target_quad = np.transpose(
        np.vstack(
          [
            0.5 * ctx_Jaref_np * ctx_Jaref_np * efc_D_np,
            target_jv * ctx_Jaref_np * efc_D_np,
            0.5 * target_jv * target_jv * efc_D_np,
          ]
        )
      )

      # Reset and launch linesearch
      ctx.jv.zero_()
      ctx.quad.zero_()
      ctx.quad_gauss.zero_()
      step_size_cost = wp.empty((d.nworld, m.opt.ls_iterations), dtype=float)
      solver._linesearch(m, d, ctx, step_size_cost)

      # mv and jv are always written
      ctx_mv = ctx.mv.numpy()[0]
      ctx_jv = ctx.jv.numpy()[0]
      _assert_eq(ctx_mv, target_mv, "mv")
      _assert_eq(ctx_jv[:nefc], target_jv[:nefc], "jv")

      if ls_parallel and cone == ConeType.PYRAMIDAL:
        # Parallel pyramidal has separate prep kernels that write quad_gauss and quad
        # (Elliptic quad uses special quad1/quad2 format that target_quad doesn't compute)
        ctx_quad_gauss = ctx.quad_gauss.numpy()[0]
        ctx_quad = ctx.quad.numpy()[0]
        _assert_eq(ctx_quad_gauss, target_quad_gauss, "quad_gauss")
        _assert_eq(ctx_quad[:nefc], target_quad[:nefc], "quad")
      elif ls_parallel and cone == ConeType.ELLIPTIC:
        # Parallel elliptic: only check quad_gauss (quad uses special format)
        ctx_quad_gauss = ctx.quad_gauss.numpy()[0]
        _assert_eq(ctx_quad_gauss, target_quad_gauss, "quad_gauss")

  @parameterized.product(
    cone=(ConeType.PYRAMIDAL, ConeType.ELLIPTIC), jacobian=(mujoco.mjtJacobian.mjJAC_SPARSE, mujoco.mjtJacobian.mjJAC_DENSE)
  )
  def test_update_gradient_CG(self, cone, jacobian):
    """Test _update_gradient function is correct for the CG solver."""
    mjm, mjd, m, d = test_data.fixture(
      "humanoid/humanoid.xml",
      keyframe=0,
      overrides={"opt.cone": cone, "opt.solver": SolverType.CG, "opt.jacobian": jacobian, "opt.iterations": 0},
    )

    # Create SolverContext and initialize
    ctx = solver.create_solver_context(m, d)
    solver.init_context(m, d, ctx, grad=True)

    # Calculate Mgrad with Mujoco C
    mj_Mgrad = np.zeros(shape=(1, mjm.nv), dtype=float)
    mj_grad = np.tile(ctx.grad.numpy()[:, : mjm.nv], (1, 1))
    mujoco.mj_solveM(mjm, mjd, mj_Mgrad, mj_grad)

    ctx_Mgrad = ctx.Mgrad.numpy()[0, : mjm.nv]
    _assert_eq(ctx_Mgrad, mj_Mgrad[0], name="Mgrad")

  @parameterized.parameters(ConeType.PYRAMIDAL, ConeType.ELLIPTIC)
  def test_parallel_linesearch(self, cone):
    """Test that iterative and parallel linesearch leads to equivalent results."""
    _, _, m, d = test_data.fixture(
      "humanoid/humanoid.xml",
      qpos_noise=0.01,
      overrides={"opt.cone": cone, "opt.iterations": 50, "opt.ls_iterations": 50},
    )

    # One step to obtain more non-zeros results
    mjw.step(m, d)

    # Preparing for linesearch
    m.opt.iterations = 0
    mjw.fwd_velocity(m, d)
    mjw.fwd_acceleration(m, d, factorize=True)
    ctx = solver.create_solver_context(m, d)
    solver._solve(m, d, ctx)

    # Storing some initial values
    d_efc_Ma = d.efc.Ma.numpy().copy()
    ctx_Jaref = ctx.Jaref.numpy().copy()
    d_qacc = d.qacc.numpy().copy()

    # Launching iterative linesearch
    m.opt.ls_parallel = False
    step_size_cost = wp.empty((d.nworld, 0), dtype=float)
    solver._linesearch(m, d, ctx, step_size_cost)
    # Iterative computes alpha internally and directly updates outputs
    qacc_iterative = d.qacc.numpy().copy()
    Ma_iterative = d.efc.Ma.numpy().copy()
    Jaref_iterative = ctx.Jaref.numpy().copy()

    # Launching parallel linesearch with 50 testing points
    m.opt.ls_parallel = True
    m.opt.ls_iterations = 50
    d.efc.Ma = wp.array2d(d_efc_Ma)
    ctx.Jaref = wp.array2d(ctx_Jaref)
    d.qacc = wp.array2d(d_qacc)
    step_size_cost = wp.empty((d.nworld, m.opt.ls_iterations), dtype=float)
    solver._linesearch(m, d, ctx, step_size_cost)
    qacc_parallel = d.qacc.numpy().copy()
    Ma_parallel = d.efc.Ma.numpy().copy()
    Jaref_parallel = ctx.Jaref.numpy().copy()

    # Check that iterative and parallel linesearch produce equivalent outputs
    _assert_eq(qacc_iterative, qacc_parallel, name="qacc")
    _assert_eq(Ma_iterative, Ma_parallel, name="Ma")
    _assert_eq(Jaref_iterative, Jaref_parallel, name="Jaref")

  @parameterized.parameters(
    (ConeType.PYRAMIDAL, SolverType.CG, 10, 5, mujoco.mjtJacobian.mjJAC_DENSE, False),
    (ConeType.ELLIPTIC, SolverType.CG, 10, 5, mujoco.mjtJacobian.mjJAC_DENSE, False),
    (ConeType.PYRAMIDAL, SolverType.NEWTON, 5, 10, mujoco.mjtJacobian.mjJAC_DENSE, False),
    (ConeType.ELLIPTIC, SolverType.NEWTON, 5, 10, mujoco.mjtJacobian.mjJAC_DENSE, False),
    (ConeType.PYRAMIDAL, SolverType.NEWTON, 5, 64, mujoco.mjtJacobian.mjJAC_SPARSE, True),
    (ConeType.ELLIPTIC, SolverType.NEWTON, 5, 64, mujoco.mjtJacobian.mjJAC_SPARSE, True),
  )
  def test_solve(self, cone, solver_, iterations, ls_iterations, jacobian, ls_parallel):
    """Tests solve."""
    for keyframe in range(3):
      mjm, mjd, m, d = test_data.fixture(
        "constraints.xml",
        keyframe=keyframe,
        overrides={
          "opt.jacobian": jacobian,
          "opt.cone": cone,
          "opt.solver": solver_,
          "opt.iterations": iterations,
          "opt.ls_iterations": ls_iterations,
          "opt.ls_parallel": ls_parallel,
        },
      )

      mujoco.mj_forward(mjm, mjd)

      d.qacc.zero_()
      d.qfrc_constraint.zero_()
      d.efc.force.zero_()

      if solver_ == mujoco.mjtSolver.mjSOL_CG:
        mjw.factor_m(m, d)
      mjw.solve(m, d)

      def cost(qacc):
        jaref = np.zeros(mjd.nefc, dtype=float)
        cost = np.zeros(1)
        mujoco.mj_mulJacVec(mjm, mjd, jaref, qacc)
        mujoco.mj_constraintUpdate(mjm, mjd, jaref - mjd.efc_aref, cost, 0)
        return cost

      mj_cost = cost(mjd.qacc)
      mjwarp_cost = cost(d.qacc.numpy()[0])
      self.assertLessEqual(mjwarp_cost, mj_cost * 1.025)

      if m.opt.solver == mujoco.mjtSolver.mjSOL_NEWTON:
        _assert_eq(d.qacc.numpy()[0], mjd.qacc, "qacc")
        _assert_eq(d.qfrc_constraint.numpy()[0], mjd.qfrc_constraint, "qfrc_constraint")
        _assert_eq(d.efc.force.numpy()[0, : mjd.nefc], mjd.efc_force, "efc_force")

  @parameterized.parameters(
    (ConeType.PYRAMIDAL, SolverType.CG, 25, 5),
    (ConeType.PYRAMIDAL, SolverType.NEWTON, 2, 4),
  )
  def test_solve_batch(self, cone, solver_, iterations, ls_iterations):
    """Tests solve (batch)."""
    mjm0, mjd0, _, _ = test_data.fixture(
      "humanoid/humanoid.xml",
      keyframe=0,
      overrides={"opt.cone": cone, "opt.solver": solver_, "opt.iterations": iterations, "opt.ls_iterations": ls_iterations},
    )
    qacc_warmstart0 = mjd0.qacc_warmstart.copy()
    mujoco.mj_forward(mjm0, mjd0)
    mjd0.qacc_warmstart = qacc_warmstart0

    mjm1, mjd1, _, _ = test_data.fixture(
      "humanoid/humanoid.xml",
      keyframe=2,
      overrides={"opt.cone": cone, "opt.solver": solver_, "opt.iterations": iterations, "opt.ls_iterations": ls_iterations},
    )
    qacc_warmstart1 = mjd1.qacc_warmstart.copy()
    mujoco.mj_forward(mjm1, mjd1)
    mjd1.qacc_warmstart = qacc_warmstart1

    mjm2, mjd2, _, _ = test_data.fixture(
      "humanoid/humanoid.xml",
      keyframe=1,
      overrides={"opt.cone": cone, "opt.solver": solver_, "opt.iterations": iterations, "opt.ls_iterations": ls_iterations},
    )
    qacc_warmstart2 = mjd2.qacc_warmstart.copy()
    mujoco.mj_forward(mjm2, mjd2)
    mjd2.qacc_warmstart = qacc_warmstart2

    nefc_active = mjd0.nefc + mjd1.nefc + mjd2.nefc
    ne_active = mjd0.ne + mjd1.ne + mjd2.ne

    mjm, mjd, m, _ = test_data.fixture(
      "humanoid/humanoid.xml",
      overrides={"opt.cone": cone, "opt.solver": solver_, "opt.iterations": iterations, "opt.ls_iterations": ls_iterations},
    )
    d = mjw.put_data(mjm, mjd, nworld=3, njmax=2 * nefc_active)

    d.nefc = wp.array([nefc_active, nefc_active, nefc_active], dtype=wp.int32, ndim=1)
    d.ne = wp.array([ne_active, ne_active, ne_active], dtype=wp.int32, ndim=1)

    qacc_warmstart = np.vstack(
      [
        np.expand_dims(qacc_warmstart0, axis=0),
        np.expand_dims(qacc_warmstart1, axis=0),
        np.expand_dims(qacc_warmstart2, axis=0),
      ]
    )

    qM0 = np.zeros((mjm0.nv, mjm0.nv))
    mujoco.mj_fullM(mjm0, qM0, mjd0.qM)
    qM1 = np.zeros((mjm1.nv, mjm1.nv))
    mujoco.mj_fullM(mjm1, qM1, mjd1.qM)
    qM2 = np.zeros((mjm2.nv, mjm2.nv))
    mujoco.mj_fullM(mjm2, qM2, mjd2.qM)

    qM = np.vstack(
      [
        np.expand_dims(qM0, axis=0),
        np.expand_dims(qM1, axis=0),
        np.expand_dims(qM2, axis=0),
      ]
    )
    qacc_smooth = np.vstack(
      [
        np.expand_dims(mjd0.qacc_smooth, axis=0),
        np.expand_dims(mjd1.qacc_smooth, axis=0),
        np.expand_dims(mjd2.qacc_smooth, axis=0),
      ]
    )
    qfrc_smooth = np.vstack(
      [
        np.expand_dims(mjd0.qfrc_smooth, axis=0),
        np.expand_dims(mjd1.qfrc_smooth, axis=0),
        np.expand_dims(mjd2.qfrc_smooth, axis=0),
      ]
    )

    # Reshape the Jacobians
    efc_J0 = mjd0.efc_J.reshape((mjd0.nefc, mjm0.nv))
    efc_J1 = mjd1.efc_J.reshape((mjd1.nefc, mjm1.nv))
    efc_J2 = mjd2.efc_J.reshape((mjd2.nefc, mjm2.nv))

    efc_J_fill = np.zeros((3, d.njmax, m.nv))
    efc_J_fill[0, : mjd0.nefc, :] = efc_J0
    efc_J_fill[1, : mjd1.nefc, :] = efc_J1
    efc_J_fill[2, : mjd2.nefc, :] = efc_J2

    # Similarly for D and aref values
    efc_D0 = mjd0.efc_D[: mjd0.nefc]
    efc_D1 = mjd1.efc_D[: mjd1.nefc]
    efc_D2 = mjd2.efc_D[: mjd2.nefc]

    efc_D_fill = np.zeros((3, d.njmax))
    efc_D_fill[0, : mjd0.nefc] = efc_D0
    efc_D_fill[1, : mjd1.nefc] = efc_D1
    efc_D_fill[2, : mjd2.nefc] = efc_D2

    efc_aref0 = mjd0.efc_aref[: mjd0.nefc]
    efc_aref1 = mjd1.efc_aref[: mjd1.nefc]
    efc_aref2 = mjd2.efc_aref[: mjd2.nefc]

    efc_aref_fill = np.zeros((3, d.njmax))
    efc_aref_fill[0, : mjd0.nefc] = efc_aref0
    efc_aref_fill[1, : mjd1.nefc] = efc_aref1
    efc_aref_fill[2, : mjd2.nefc] = efc_aref2

    d.qacc_warmstart = wp.from_numpy(qacc_warmstart, dtype=wp.float32)
    d.qM = wp.from_numpy(qM, dtype=wp.float32)
    d.qacc_smooth = wp.from_numpy(qacc_smooth, dtype=wp.float32)
    d.qfrc_smooth = wp.from_numpy(qfrc_smooth, dtype=wp.float32)
    d.efc.J = wp.from_numpy(efc_J_fill, dtype=wp.float32)
    d.efc.D = wp.from_numpy(efc_D_fill, dtype=wp.float32)
    d.efc.aref = wp.from_numpy(efc_aref_fill, dtype=wp.float32)

    if solver_ == SolverType.CG:
      m0 = mjw.put_model(mjm0)
      d0 = mjw.put_data(mjm0, mjd0)
      mjw.factor_m(m0, d0)
      qLD0 = d0.qLD.numpy()

      m1 = mjw.put_model(mjm1)
      d1 = mjw.put_data(mjm1, mjd1)
      mjw.factor_m(m1, d1)
      qLD1 = d1.qLD.numpy()

      m2 = mjw.put_model(mjm2)
      d2 = mjw.put_data(mjm2, mjd2)
      mjw.factor_m(m2, d2)
      qLD2 = d2.qLD.numpy()

      qLD = np.vstack([qLD0, qLD1, qLD2])
      d.qLD = wp.from_numpy(qLD, dtype=wp.float32)

    d.qacc.zero_()
    d.qfrc_constraint.zero_()
    d.efc.force.zero_()
    solver.solve(m, d)

    def cost(m, d, qacc):
      jaref = np.zeros(d.nefc, dtype=float)
      cost = np.zeros(1)
      mujoco.mj_mulJacVec(m, d, jaref, qacc)
      mujoco.mj_constraintUpdate(m, d, jaref - d.efc_aref, cost, 0)
      return cost

    mj_cost0 = cost(mjm0, mjd0, mjd0.qacc)
    mjwarp_cost0 = cost(mjm0, mjd0, d.qacc.numpy()[0])
    self.assertLessEqual(mjwarp_cost0, mj_cost0 * 1.025)

    mj_cost1 = cost(mjm1, mjd1, mjd1.qacc)
    mjwarp_cost1 = cost(mjm1, mjd1, d.qacc.numpy()[1])
    self.assertLessEqual(mjwarp_cost1, mj_cost1 * 1.025)

    mj_cost2 = cost(mjm2, mjd2, mjd2.qacc)
    mjwarp_cost2 = cost(mjm2, mjd2, d.qacc.numpy()[2])
    self.assertLessEqual(mjwarp_cost2, mj_cost2 * 1.025)

    if m.opt.solver == SolverType.NEWTON:
      _assert_eq(d.qacc.numpy()[0], mjd0.qacc, "qacc0")
      _assert_eq(d.qacc.numpy()[1], mjd1.qacc, "qacc1")
      _assert_eq(d.qacc.numpy()[2], mjd2.qacc, "qacc2")

      _assert_eq(d.qfrc_constraint.numpy()[0], mjd0.qfrc_constraint, "qfrc_constraint0")
      _assert_eq(d.qfrc_constraint.numpy()[1], mjd1.qfrc_constraint, "qfrc_constraint1")
      _assert_eq(d.qfrc_constraint.numpy()[2], mjd2.qfrc_constraint, "qfrc_constraint2")

      # Get world 0 forces - equality constraints at start, inequality constraints later
      nieq0 = mjd0.nefc - mjd0.ne
      nieq1 = mjd1.nefc - mjd1.ne
      nieq2 = mjd2.nefc - mjd2.ne
      world0_eq_forces = d.efc.force.numpy()[0, : mjd0.ne]
      world0_ineq_forces = d.efc.force.numpy()[0, ne_active : ne_active + nieq0]
      world0_forces = np.concatenate([world0_eq_forces, world0_ineq_forces])
      _assert_eq(world0_forces, mjd0.efc_force, "efc_force0")

      # Get world 1 forces
      world1_eq_forces = d.efc.force.numpy()[1, : mjd1.ne]
      world1_ineq_forces = d.efc.force.numpy()[1, ne_active : ne_active + nieq1]
      world1_forces = np.concatenate([world1_eq_forces, world1_ineq_forces])
      _assert_eq(world1_forces, mjd1.efc_force, "efc_force1")

      # Get world 2 forces
      world2_eq_forces = d.efc.force.numpy()[2, : mjd2.ne]
      world2_ineq_forces = d.efc.force.numpy()[2, ne_active : ne_active + nieq2]
      world2_forces = np.concatenate([world2_eq_forces, world2_ineq_forces])
      _assert_eq(world2_forces, mjd2.efc_force, "efc_force2")

  def test_frictionloss(self):
    """Tests solver with frictionloss."""
    for keyframe in range(3):
      _, mjd, m, d = test_data.fixture("constraints.xml", keyframe=keyframe)
      mjw.solve(m, d)

      _assert_eq(d.nf.numpy()[0], mjd.nf, "nf")
      _assert_eq(d.qacc.numpy()[0], mjd.qacc, "qacc")
      _assert_eq(d.qfrc_constraint.numpy()[0], mjd.qfrc_constraint, "qfrc_constraint")
      _assert_eq(d.efc.force.numpy()[0, : mjd.nefc], mjd.efc_force, "efc_force")

  def test_parallel_linesearch_threads_per_efc_gt_1(self):
    """Test parallel linesearch with threads_per_efc > 1."""
    xml = """
    <mujoco>
      <worldbody>
        <body>
          <freejoint/>
          <geom size="0.1"/>
        </body>
        <body>
          <freejoint/>
          <geom size="0.1"/>
        </body>
        <body>
          <freejoint/>
          <geom size="0.1"/>
        </body>
        <body>
          <freejoint/>
          <geom size="0.1"/>
        </body>
        <body>
          <freejoint/>
          <geom size="0.1"/>
        </body>
        <body>
          <freejoint/>
          <geom size="0.1"/>
        </body>
        <body>
          <freejoint/>
          <geom size="0.1"/>
        </body>
        <body>
          <freejoint/>
          <geom size="0.1"/>
        </body>
        <body>
          <freejoint/>
          <geom size="0.1"/>
        </body>
      </worldbody>
    </mujoco>
    """
    mjm, mjd, m, d = test_data.fixture(xml=xml)
    self.assertEqual(mjm.nv, 54)  # 9 freejoints * 6 dofs each

    # parallel linesearch path with nv > 50 -> threads_per_efc > 1
    m.opt.ls_parallel = True
    m.opt.iterations = 1
    m.opt.ls_iterations = 1
    mjw.step(m, d)


if __name__ == "__main__":
  wp.init()
  absltest.main()
